from __future__ import annotations

from collections import defaultdict
from contextlib import AbstractContextManager
import dataclasses
from io import BufferedReader
import logging
from pathlib import Path
import sys
from textwrap import indent
import traceback
from typing import Any, Optional, Union

import construct as cs
import numpy as np
from skimage.transform import AffineTransform
from skimage.transform import warp

from eyepy.core.eyeenface import EyeEnface
from eyepy.core.eyemeta import EyeBscanMeta
from eyepy.core.eyemeta import EyeEnfaceMeta
from eyepy.core.eyemeta import EyeVolumeMeta
from eyepy.core.eyevolume import EyeVolume
from eyepy.io.utils import _compute_localizer_oct_transform
from eyepy.io.utils import find_float
from eyepy.io.utils import find_int
from eyepy.io.utils import get_bscan_spacing

from .e2e_format import ContainerHeader
from .e2e_format import containerheader_format
from .e2e_format import DataContainer
from .e2e_format import datacontainer_format
from .e2e_format import e2e_format
from .e2e_format import Type10025
from .e2e_format import TypesEnum
from .vol_reader import SEG_MAPPING

logger = logging.getLogger(__name__)

# Occurence of type ids. This is used by the inspect function.
type_occurence = {
    'E2EFileStructure': [0, 9011],
    'E2EPatientStructure': [9, 17, 29, 31, 52, 9010],
    'E2EStudyStructure': [7, 10, 13, 30, 53, 58, 1000, 9000, 9001],
    'E2ESeriesStructure': [
        2, 3, 11, 54, 59, 61, 62, 1000, 1001, 1003, 1008, 9005, 9006, 9007,
        9008, 10005, 10009, 10010, 10011, 10013, 10025, 1073741824, 1073751824,
        1073751825, 1073751826
    ],
    'E2ESliceStructure': [
        2, 3, 5, 39, 40, 10004, 10012, 10013, 10019, 10020, 10032, 1073741824,
        1073751824, 1073751825, 1073751826
    ]
}


class E2EStructureMixin:
    """A Mixin for shared functionality between structures in the E2E
    hierarchy."""

    def inspect(self,
                recursive: bool = False,
                ind_prefix: str = '',
                tables: bool = False) -> str:
        """Inspect the E2E structure.

        Args:
            recursive: If True inspect lower level structures recursively.
            ind_prefix: Indentation for showing information from lower level structures.
            tables: If True add markdown table overview of the contained folder types.

        Returns:
            Information about the E2E structure.
        """
        text = self._get_section_title() + '\n'
        text += self._get_section_description() + '\n'
        if tables:
            text += self._get_folder_summary() + '\n'

        if not recursive:
            return text

        for s in self.substructure.values():
            text += '\n'
            text += indent(s.inspect(recursive, ind_prefix, tables),
                           ind_prefix)
        return text

    def get_folder_data(
        self,
        folder_type: Union[TypesEnum, int],
        offset: int = 0,
        data_construct: Optional[Union[cs.Construct, str]] = None,
    ) -> Any:
        """Return the data of a folder type.

        Args:
            folder_type: Either one of [TypesEnum][eyepy.io.he.e2e_format.TypesEnum] or the type id (int).
            offset: Offset to the data in bytes.
            data_construct: A construct to parse the data with (Python construct package) or a string describing one of the basic constructs from the construct package like "Int8ul", or "Float32l"

        Returns:
            Parsed data or None if no folder of the given type was found.
        """

        folders: list[E2EFolder] = self.folders[folder_type]

        if len(folders) == 0:
            return None

        if data_construct is None:
            return [f.data for f in folders]
        elif type(data_construct) == str:
            data_construct = getattr(cs, data_construct)

        return [f.parse_spec(data_construct, offset) for f in folders]

    def __str__(self):
        return self.inspect()

    def _get_section_title(self) -> str:
        """Make a title for describing the structure.

        Used by the inspect function.
        """
        if not self._section_title:
            try:
                self._section_title = f'{self.__class__.__name__}({self.id})'
            except:
                self._section_title = f'{self.__class__.__name__}'

        return self._section_title

    def _get_table(self, data, structure=None) -> str:
        """Make a markdown table.

        Used by the inspect function.
        """
        if structure is None:
            structure = self.__class__.__name__
        data = [[
            f'{TypesEnum(k).name} ({k})' if k in TypesEnum else f'{k}',
            len(v),
            np.mean(v),
            np.min(v),
            np.max(v), False if k not in type_occurence[structure] else True
        ] for k, v in data.items()]
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                'pandas is required for table output. Please install pandas or use the inspect function without tables=True.')
        text = pd.DataFrame.from_records(data,
                                         columns=[
                                             'Type', 'Count', 'Mean Size',
                                             'Min Size', 'Max Size',
                                             'described'
                                         ]).to_markdown(index=False)
        return text

    def _get_folder_summary(self) -> str:
        """Make a markdown table with folder type summary for the structure.

        Used by the inspect function.
        """
        data = defaultdict(list)
        for f_list in self.folders.values():
            for f in f_list:
                data[f.type].append(f.size)

        text = self._get_table(data)
        return text

    def _get_section_description(self) -> str:
        """Make a description for describing the structure.

        This uses the _section_description_parts attribute that can be
        defined in a structure to make a description.

        Used by the inspect function.
        """
        if not self._section_description:
            self._section_description = ''
            for part in self._section_description_parts:
                try:
                    self._section_description += f'{part[0]} {self.folders[part[1]][0].data.text[part[2]]} - '
                except:
                    pass
            self._section_description = self._section_description.rstrip(' - ')
        return self._section_description


@dataclasses.dataclass
class E2EFolder():
    """Folder data class.

    !!! Note

        Folders are created during initialization of the HeE2eReader. For accessing the data the
        respective HeE2eReader has to be used as a Context Manager. This opens the E2E file and
        allows the E2EFolder to access the data.


        ```python
        with HeE2eReader("path/to/e2e") as reader:
            folder_dict = reader.file_hierarchy.folders
            folder = folder_dict[TypesEnum.YOURTYPE][0]
            data = folder.data
        ```
    """
    patient_id: int
    study_id: int
    series_id: int
    slice_id: int
    pos: int
    start: int
    type: int
    size: int
    ind: int
    reader: 'HeE2eReader'

    _data = None
    _header = None

    @property
    def file_object(self) -> BufferedReader:
        """Return the file object.

        This refers to the the HeE2eReader file object.
        """
        return self.reader.file_object

    @property
    def data(self) -> Any:
        """Return the data."""
        if not self._data:
            parsed = self._parse_data()
            self._data = parsed.item
            self._header = parsed.header
        return self._data

    @property
    def header(self) -> ContainerHeader:
        """Return the data header."""
        if not self._header:
            parsed = self._parse_data()
            self._data = parsed.item
            self._header = parsed.header
        return self._header

    def _parse_data(self) -> DataContainer:
        """Parse the data.

        This only works if the HeE2eReader is used as a Context Manager
        or during initialization of the HeE2eReader. Otherwise the E2E
        file is not open.
        """
        self.file_object.seek(self.start)
        return datacontainer_format.parse_stream(self.file_object)

    def parse_spec(self, data_construct: cs.Construct, offset: int = 0) -> Any:
        """Parse a data specification.

        This only works if the HeE2eReader is used as a Context Manager or during initialization of the HeE2eReader.
        Otherwise the E2E file is not open.



        Args:
            data_construct: The construct to parse the data with. You can Constructs defined in the construct library or those defined in the [e2e_format][eyepy.io.he.e2e_format] module.
            offset: The offset in bytes, 0 by default.
        """
        b = self.get_bytes()
        return data_construct.parse(b[offset:])

    def get_bytes(self) -> bytes:
        """Return the bytes of the data.

        This only works if the HeE2eReader is used as a Context Manager
        or during initialization of the HeE2eReader. Otherwise the E2E
        file is not open.
        """
        self.file_object.seek(self.start + containerheader_format.sizeof())
        return self.file_object.read(self.size)


class E2ESliceStructure(E2EStructureMixin):
    """E2E Slice Structure.

    This structure contains folders with data for a single Slice/B-csan
    and provide convenience functions for accessing the data.
    """

    def __init__(self, id: int) -> None:
        self.id = id
        self.folders: dict[Union[int, str], list[E2EFolder]] = {}

        # Empty so inspect() does not fail
        self.substructure = {}

    def add_folder(self, folder: E2EFolder) -> None:
        """Add a folder to the slice.

        Args:
            folder: The folder to add.
        """
        try:
            self.folders[folder.type].append(folder)
        except KeyError:
            self.folders[folder.type] = [folder]

    def get_layers(self) -> dict[int, np.ndarray]:
        """Return the layers as a dictionary of layer id and layer data."""
        layers = {}
        for layer_folder in self.folders[TypesEnum.layer_annotation]:
            layers[layer_folder.data.id] = layer_folder.data.data
        return layers

    def get_meta(self) -> EyeBscanMeta:
        """Return the slice meta data."""
        if len(self.folders[TypesEnum.bscanmeta]) > 1:
            logger.warning(
                'There is more than one bscanmeta object. This is not expected.'
            )


        meta = self.folders[TypesEnum.bscanmeta][0].data

        return EyeBscanMeta(  #quality=meta.quality,
            start_pos=((meta['start_x']),
                       (meta['start_y'])),
            end_pos=((meta['end_x']),
                     (meta['end_y'])),
            pos_unit='°',
            **dataclasses.asdict(meta))

    def get_bscan(self) -> np.ndarray:
        """Return the slice image (B-scan)"""
        bscan_folders = [
            f for f in self.folders[TypesEnum.image] if f.data.type == 35652097
        ]
        if len(bscan_folders) > 1:
            logger.warning(
                'There is more than one B-scan per slice. This is not expected.'
            )
        return bscan_folders[0].data.data

    def get_localizer(self) -> np.ndarray:
        """Return the slice image (Localizer/Fundus) For the scanpattern "OCT
        Bscan" a localizer might be stored in the E2ESliceStructure and not the
        E2ESeriesStructure."""
        localizer_folders = [
            f for f in self.folders[TypesEnum.image] if f.data.type == 33620481
        ]
        if len(localizer_folders) > 1:
            logger.warning(
                'There is more than one localizer per slice. This is not expected.'
            )
        return localizer_folders[0].data.data


class E2ESeriesStructure(E2EStructureMixin):
    """E2E Series Structure.

    This structure contains folders with data for a single Series/OCT-
    Volume and provides convenience functions for accessing the data.
    """

    def __init__(self, id: int) -> None:
        self.id = id
        self.substructure: dict[int, E2ESliceStructure] = {}
        self.folders: dict[Union[int, str], list[E2EFolder]] = {}

        self._meta = None
        self._bscan_meta = None
        self._localizer_meta = None
        self._section_title = ''
        self._section_description = ''

        # Description used in inspect()
        # Parts are (name, folder_id, index in list of strings)
        self._section_description_parts = [
            ('Structure:', 9005, 0),
            ('Scanpattern:', 9006, 0),
            ('Oct Modality:', 9008, 1),
            ('Enface Modality:', 9007, 1),
        ]

    def add_folder(self, folder: E2EFolder) -> None:
        """Add a folder to the Series.

        Args:
            folder: The folder to add.
        """
        if folder.slice_id == -1:
            try:
                self.folders[folder.type].append(folder)
            except KeyError:
                self.folders[folder.type] = [folder]
        else:
            if folder.slice_id not in self.slices:
                self.slices[folder.slice_id] = E2ESliceStructure(
                    folder.slice_id)
            self.slices[folder.slice_id].add_folder(folder)

    def inspect(self,
                recursive: bool = False,
                ind_prefix: str = '',
                tables: bool = False) -> str:
        """Inspect the series.

        Custom `inspect` method to print a summary table for the slices belonging to the series.

        Args:
            recursive: If True inspect lower level structures recursively.
            ind_prefix: Indentation for showing information from lower level structures.
            tables: If True add markdown table overview of the contained folder types.
        """
        laterality = self.folders[TypesEnum.laterality][
            0].data.laterality.name if TypesEnum.laterality in self.folders else 'Unknown'
        text = self._get_section_title(
        ) + f' - Laterality: {laterality} - B-scans: {self.n_bscans}\n'
        text += self._get_section_description() + '\n'
        if tables:
            text += self._get_folder_summary() + '\n'

        if not recursive:
            return text

        # Describe all slices in one table
        s_data = defaultdict(list)
        for sl in self.slices.values():
            for f_list in sl.folders.values():
                for f in f_list:
                    s_data[f.type].append(f.size)

        if len(s_data) == 0 or tables == False:
            text += ''
        else:
            text += '\nE2ESlice Summary:\n'
            text += indent(self._get_table(s_data, 'E2ESliceStructure'),
                           ind_prefix)
            text += '\n'
        return text

    def get_volume(self) -> EyeVolume:
        """Return EyeVolume object for the series."""
        ## Check if scan is a volume scan
        volume_meta = self.get_meta()

        scan_pattern = volume_meta['bscan_meta'][0]['scan_pattern']

        ## Check if scan pattern is supported by EyeVolume
        if scan_pattern == 2:
            msg = f'The EyeVolume object does not support scan pattern 2 (one Circular B-scan).'
            raise ValueError(msg)
        elif scan_pattern == 5:
            msg = f'The EyeVolume object does not support scan pattern 5 (Radial scan - star pattern).'
            raise ValueError(msg)

        data = self.get_bscans()

        volume_meta = self.get_meta()
        localizer = self.get_localizer()
        volume = EyeVolume(
            data=data,
            meta=volume_meta,
            localizer=localizer,
            transformation=_compute_localizer_oct_transform(
                volume_meta, localizer.meta, data.shape),
        )

        layer_height_maps = self.get_layers()
        for name, i in SEG_MAPPING.items():
            if i in layer_height_maps:
                volume.add_layer_annotation(layer_height_maps[i], name=name)

        return volume

    @property
    def n_bscans(self) -> int:
        """Return the number of B-scans in the series."""
        return len(self.substructure)

    def get_bscans(self) -> np.ndarray:
        volume_meta = self.get_meta()
        size_x = volume_meta['bscan_meta'][0]['size_x']
        size_y = volume_meta['bscan_meta'][0]['size_y']

        data = np.zeros((self.n_bscans, size_y, size_x))
        for ind, sl in self.slices.items():
            bscan = sl.get_bscan()
            i = ind // 2 if len(
                self.get_bscan_meta()
            ) != 1 else 0  # Slice id for single B-scan Volumes is 2 and not 0 in the e2e file.
            data[i] = bscan
        return data

    def get_layers(self) -> dict[int, np.ndarray]:
        """Return layer height maps for the series as dict of numpy arrays where
        the key is the layer id."""
        slice_layers = {}
        layer_ids = set()

        for ind, sl in self.slices.items():
            layers = sl.get_layers()
            [layer_ids.add(k) for k in layers.keys()]
            slice_layers[ind // 2] = layers

        layers = {}
        size_x = self.get_bscan_meta()[0]['size_x']
        for i in layer_ids:
            layer = np.full((self.n_bscans, size_x), np.nan)
            if self.n_bscans == 1:
                layer[0, :] = slice_layers[1][i]
                layers[i] = layer

            else:
                for sl in range(self.n_bscans):
                    layer[sl, :] = slice_layers[sl][i]

            layer[layer >= 3.0e+38] = np.nan
            layers[i] = layer

        return layers

    def enface_modality(self) -> str:
        folders = self.folders[TypesEnum.enface_modality]
        if len(folders) > 1:
            logger.debug(
                'There is more than one enface modality stored. This is not expected.'
            )
        text = folders[0].data.text[1]
        return 'NIR' if text == 'IR' else text

    def laterality(self) -> str:
        folders = self.folders[TypesEnum.laterality]
        if len(folders) > 1:
            logger.debug(
                'There is more than one laterality stored. This is not expected.'
            )
        return str(folders[0].data.laterality)

    def slo_data(self) -> Type10025:
        folders = self.folders[TypesEnum.slodata]
        if len(folders) > 1:
            logger.debug(
                'There is more than one SLO data folder. This is not expected.'
            )
        return folders[0].data

    def localizer_meta(self, height, width) -> EyeEnfaceMeta:
        """Return EyeEnfaceMeta object for the localizer image."""
        if self._localizer_meta is None:
            self._localizer_meta = EyeEnfaceMeta(
                scale_x=30 / width,  # Give scale in degrees per pixel
                scale_y=30 / height,
                scale_unit='°',
                modality=self.enface_modality(),
                laterality=self.laterality(),
                field_size=None,
                scan_focus=None,
                visit_date=None,
                exam_time=None,
            )
        logger.info(
            'The localizer scale is currently hardcoded and not read from the E2E file. If you know how or where to find the scale information let us know by opening an issue.'
        )
        return self._localizer_meta

    def get_localizer(self) -> EyeEnface:
        """Return EyeEnface object for the localizer image."""
        try:
            folders = self.folders[TypesEnum.image]
            if len(folders) > 1:
                logger.warning(
                    'There is more than one enface localizer image stored. This is not expected.'
                )

            # Slodata is not always present in E2E files.
            # Todo: Give transform to EyeEnface object where it is applied to the image. EyeEnface then by default has an identity transform.
            #transform = np.array(list(self.slo_data().transform) +
            #                     [0, 0, 1]).reshape((3, 3))
            # transfrom localizer with transform from E2E file
            #transformed_localizer = warp(folders[0].data.data,
            #                             AffineTransform(transform),
            #                             order=1,
            #                             preserve_range=True)
            return EyeEnface(folders[0].data.data,
                             self.localizer_meta(height=folders[0].data.height,
                                                 width=folders[0].data.width))
        except KeyError:
            if self.n_bscans == 1:
                slice_struct = self.slices[2]
                return EyeEnface(slice_struct.get_localizer(),
                                 self.localizer_meta())
            else:
                raise ValueError(
                    'There is no localizer/fundus image in the E2E file.')

    def get_bscan_meta(self) -> list[EyeBscanMeta]:
        """Return EyeBscanMeta objects for all B-scans in the series."""
        if self._bscan_meta is None:
            self._bscan_meta = sorted(
                [sl.get_meta() for sl in self.slices.values()],
                key=lambda x: x['aktImage'])
        return self._bscan_meta

    def get_meta(self) -> EyeVolumeMeta:
        """Return EyeVolumeMeta object for the series."""
        if self._meta is None:
            bscan_meta = self.get_bscan_meta()
            self._meta = EyeVolumeMeta(
                scale_x=1,  #0.0114,  # Todo: Where is this in E2E?
                scale_y=1,  #bscan_meta[0]["scale_y"],
                scale_z=1,  #get_bscan_spacing(bscan_meta) if
                #(bscan_meta[0]["scan_pattern"] not in [1, 2]) else 0.03,
                scale_unit='px',
                laterality=self.laterality(),
                visit_date=None,
                exam_time=None,
                bscan_meta=bscan_meta,
                intensity_transform='e2e',
            )
        return self._meta

    @property
    def slices(self) -> dict[int, E2ESliceStructure]:
        """Alias for substructure."""
        return self.substructure


class E2EStudyStructure(E2EStructureMixin):
    """E2E Study Structure."""

    def __init__(self, id) -> None:
        self.id = id
        self.substructure: dict[int, E2ESeriesStructure] = {}
        self.folders: dict[Union[int, str], list[E2EFolder]] = {}

        self._section_description_parts = [('Device:', 9001, 0),
                                           ('Studyname:', 9000, 0)]
        self._section_title = ''
        self._section_description = ''

    @property
    def series(self) -> dict[int, E2ESeriesStructure]:
        return self.substructure

    def add_folder(self, folder: E2EFolder) -> None:
        """Add a folder to the Study.

        Args:
            folder: The folder to add.
        """
        if folder.series_id == -1:
            try:
                self.folders[folder.type].append(folder)
            except KeyError:
                self.folders[folder.type] = [folder]
        else:
            if folder.series_id not in self.series:
                self.series[folder.series_id] = E2ESeriesStructure(
                    folder.series_id)
            self.series[folder.series_id].add_folder(folder)


class E2EPatientStructure(E2EStructureMixin):
    """E2E Patient Structure."""

    def __init__(self, id) -> None:
        self.id = id
        self.substructure: dict[int, E2EStudyStructure] = {}
        self.folders: dict[Union[int, str], list[E2EFolder]] = {}

        self._section_description_parts = []
        self._section_title = ''
        self._section_description = ''

    @property
    def studies(self) -> dict[int, E2EStudyStructure]:
        return self.substructure

    def add_folder(self, folder: E2EFolder) -> None:
        """Add a folder to the Patient Structure.

        Args:
            folder: The folder to add.
        """
        if folder.study_id == -1:
            try:
                self.folders[folder.type].append(folder)
            except KeyError:
                self.folders[folder.type] = [folder]
        else:
            if folder.study_id not in self.studies:
                self.studies[folder.study_id] = E2EStudyStructure(
                    folder.study_id)
            self.studies[folder.study_id].add_folder(folder)


class E2EFileStructure(E2EStructureMixin):
    """E2E File Structure."""

    def __init__(self):
        self.substructure: dict[int, E2EPatientStructure] = {}
        self.folders: dict[Union[int, str], list[E2EFolder]] = {}

        self._section_description_parts = []
        self._section_title = ''
        self._section_description = ''

    @property
    def patients(self) -> dict[int, E2EPatientStructure]:
        return self.substructure

    def add_folder(self, folder: E2EFolder):
        """Add a folder to the File Structure.

        Args:
            folder: The folder to add.
        """
        try:
            self.all_folders.append(folder)
        except AttributeError:
            self.all_folders = [folder]

        if folder.patient_id == -1:
            try:
                self.folders[folder.type].append(folder)
            except KeyError:
                self.folders[folder.type] = [folder]
        else:
            if folder.patient_id not in self.patients:
                self.patients[folder.patient_id] = E2EPatientStructure(
                    folder.patient_id)
            self.patients[folder.patient_id].add_folder(folder)


class HeE2eReader(AbstractContextManager):

    def __init__(self, path: Union[str, Path]):
        """Index an E2E file.

        Initialization of the HeE2eReader class indexes the specified E2E file. This allows for printing the reader object
        for a quick overview of the files contents. If you want to access the data, the reader has to be used as a Context Manager.

        ```python
        with HeE2eReader("path/to/file.e2e") as reader:
            data = reader.volumes
        ```

        Args:
            path: Path to the e2e file.
        """
        self.path = Path(path)
        self.file_object: BufferedReader

        # Index file to create hierarchy
        self.file_hierarchy = E2EFileStructure()
        self._index_file()

    def _index_file(self) -> None:
        with open(self.path, 'rb') as f:
            parsed = e2e_format.parse_stream(f)

            # Get the position, IDs and types of all folders
            for chunk in parsed.chunks:
                for fh in chunk.folders:
                    folder = E2EFolder(
                        **{
                            'patient_id': fh.patient_id,
                            'study_id': fh.study_id,
                            'series_id': fh.series_id,
                            'slice_id': fh.slice_id,
                            'pos': fh.pos,
                            'start': fh.start,
                            'type': fh.type,
                            'size': fh.size,
                            'ind': fh.ind,
                            'reader': self,
                        })
                    self.file_hierarchy.add_folder(folder)

            # Read and cache information required for __str__
            self.file_object = f
            self.inspect(recursive=True, ind_prefix='  ', tables=False)

    def inspect(self,
                recursive: bool = False,
                ind_prefix: str = '',
                tables: bool = True) -> str:
        """Inspect the file hierarchy (contents) of the file.

        Args:
            recursive: If True inspect lower level structures recursively.
            ind_prefix: Indentation for showing information from lower level structures.
            tables: If True add markdown table overview of the contained folder types.
        """
        return self.file_hierarchy.inspect(recursive, ind_prefix, tables)

    def __str__(self) -> str:
        return self.inspect(recursive=True, ind_prefix='  ', tables=False)

    def __repr__(self) -> str:
        return f'HeE2eReader(path="{self.path}")'

    @property
    def patients(self) -> list[E2EPatientStructure]:
        """List of all patients in the file as E2EPatient objects."""
        return [p for p in self.file_hierarchy.patients.values()]

    @property
    def studies(self) -> list[E2EStudyStructure]:
        """List of all studies in the file as E2EStudy objects."""
        studies = []
        for p in self.patients:
            studies += p.studies.values()
        return studies

    @property
    def series(self) -> list[E2ESeriesStructure]:
        """List of all series in the file as E2ESeries objects."""
        series = []
        for s in self.studies:
            series += s.series.values()
        return sorted(series, key=lambda s: s.id)

    def __enter__(self) -> 'HeE2eReader':
        self.file_object = open(self.path, 'rb')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.file_object.close()

    def find_int(self,
                 value: int,
                 excluded_folders: list[Union[int,
                                              str]] = ['images', 'layers'],
                 slice_id: Optional[int] = None,
                 **kwargs: Any) -> dict[int, dict[int, dict[str, list[int]]]]:
        """Find an integer value in the e2e file.

        Args:
            value: The value to find.
            excluded_folders: A list of folders to exclude from the search.
                None: Exclude no folders.
                "images": Exclude image data from search.
                "layers": Exclude layer data from search.
            slice_id: The slice id to search in.
            **kwargs: Keyword arguments passed to [`find_int`][eyepy.io.utils.find_int].

        Returns:
            A dictionary of the form {series_id(int): {folder_type(int): {fmt_string(str): [positions(int)]}}}
        """
        if 'images' in excluded_folders:
            excluded_folders[excluded_folders.index('images')] = 1073741824
        if 'layers' in excluded_folders:
            excluded_folders[excluded_folders.index('layers')] = 10019

        results = defaultdict(dict)
        for folder in self.file_hierarchy.all_folders:
            if not int(folder.type) in excluded_folders and (
                    True if slice_id is None else folder.slice_id == slice_id):
                res = find_int(folder.get_bytes(), value, **kwargs)
                if res:
                    results[folder.series_id][folder.type] = res
        results = {**results}
        return results

    def find_float(self,
                   value: float,
                   excluded_folders: list[Union[int,
                                                str]] = ['images', 'layers'],
                   slice_id: Optional[int] = None,
                   **kwargs: Any) -> dict[int, dict[int, dict[str, list[int]]]]:
        """Find a float value in the e2e file.

        Args:
            value: The value to find.
            excluded_folders: A list of folders to exclude from the search.
                None: Exclude no folders.
                "images": Exclude image data from search.
                "layers": Exclude layer data from search.
            slice_id: The slice to search in. Specify 0 if you do not want to search through all slices but one slice per volume is enough.
            **kwargs: Keyword arguments passed to [`find_float`][eyepy.io.utils.find_float].

        Returns:
            A dictionary of the form {series_id(int): {folder_type(int): {fmt_string(str): [positions(int)]}}}
        """
        if 'images' in excluded_folders:
            excluded_folders[excluded_folders.index('images')] = 1073741824
        if 'layers' in excluded_folders:
            excluded_folders[excluded_folders.index('layers')] = 10019

        results = defaultdict(dict)
        for folder in self.file_hierarchy.all_folders:
            if not int(folder.type) in excluded_folders and (
                    True if slice_id is None else folder.slice_id == slice_id):
                res = find_float(folder.get_bytes(), value, **kwargs)
                if res:
                    results[folder.series_id][folder.type] = res
        results = {**results}
        return results

    def find_number(self,
                    value: Union[int, float],
                    excluded_folders: list[Union[int,
                                                 str]] = ['images', 'layers'],
                    slice_id: Optional[int] = None,
                    **kwargs: Any) -> dict[int, dict[int, dict[str, list[int]]]]:
        """Find a number value in the e2e file.

        Use this function if you don't know if the value is an integer or a float.
        This is just a shortcut for calling [`find_int`][eyepy.io.he.e2e_reader.HeE2eReader.find_int]
        and [`find_float`][eyepy.io.he.e2e_reader.HeE2eReader.find_float] individually.

        Args:
            value: The value to find.
            excluded_folders: A list of folders to exclude from the search.
                None: Exclude no folders.
                "images": Exclude image data from search.
                "layers": Exclude layer data from search.
            slice_id: The slice to search in. Specify 0 if you do not want to search through all slices but one slice per volume is enough.
            **kwargs: Keyword arguments passed to [`find_int`][eyepy.io.utils.find_int] and [`find_float`][eyepy.io.utils.find_float].

        Returns:
            A dictionary of the form {series_id(int): {folder_type(int): {fmt_string(str): [positions(int)]}}}
        """
        results = {
            **self.find_float(value, excluded_folders, slice_id, **kwargs),
            **self.find_int(round(value), excluded_folders, slice_id, **kwargs)
        }
        return results

    @property
    def volume(self) -> EyeVolume:
        """First EyeVolume object in the E2E file.

        Returns:
            EyeVolume object for the first Series in the e2e file.
        """
        for s in self.series:
            try:
                return s.get_volume()
            except Exception as e:
                # for compatibility with python <= 3.9, later work with only the exception as argument for format_exception
                exc_type, exc_value, exc_tb = sys.exc_info()
                logger.debug(''.join(
                    traceback.format_exception(exc_type, exc_value, exc_tb)))
        raise ValueError(
            'No Series in the E2E file can be parsed to a an EyeVolume object. You might be able to extract information manually from the E2ESeries objects (e2ereader.series)'
        )

    @property
    def volumes(self) -> list[EyeVolume]:
        """All EyeVolume objects in the E2E file.

        Returns:
            List with EyeVolume objects for every Series in the e2e file.
        """
        volumes = []
        for s in self.series:
            try:
                volumes.append(s.get_volume())
            except Exception as e:
                logger.debug(''.join(traceback.format_exception(e)))
        return volumes



from __future__ import annotations

import cmath
from collections.abc import Sequence
import functools
import logging
from typing import Any, Iterable, Optional, TYPE_CHECKING, Union

import numpy as np
import numpy.typing as npt
from skimage import transform

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    Shape = Union[int, tuple[int, int]]


def circle_mask(radius: int,
                mask_shape: Optional[tuple[int, int]] = None,
                smooth_edges: bool = False) -> npt.NDArray[Any]:
    """Create a centered circular mask with given radius.

    Args:
        radius:
        mask_shape:
        smooth_edges:

    Returns:
    """
    if mask_shape is None:
        mask_shape = (radius * 2, radius * 2)

    if smooth_edges:
        work_shape = (mask_shape[0] * 5, mask_shape[1] * 5)
        radius *= 5
    else:
        work_shape = mask_shape

    circle_mask = np.zeros(work_shape)
    circle_mask[radius_filtergrid(
        work_shape, quadrant_shift=False, normalize=False) < radius] = 1

    return transform.resize(circle_mask, mask_shape)


def create_sectors(mask_shape: tuple[int, int],
                   n_sectors: int = 4,
                   start_angle: int = 0,
                   clockwise: bool = False,
                   smooth_edges: bool = False) -> list[npt.NDArray[Any]]:
    """Create masks for n radial sectors.

    By default the first sector is the first quadrant, and the remaining 3
    sectors are added counter clockwise.

    For a binary mask pixels can not belong to two mask without changing the
    sum over all masks. But for pixels at the sector edges it is not clear to
    which sector they belong and assigning them to two sectors partially might
    be desired. Hence if smooth_edges is True, we create 5 times bigger binary
    masks and then use the anti-aliasing from downscaling them to the desired
    shape to create a float mask.

    Args:
        mask_shape:
        n_sectors:
        start_angle:
        clockwise:
        smooth_edges:

    Returns:
    """
    if smooth_edges:
        work_shape = (mask_shape[0] * 5, mask_shape[1] * 5)
    else:
        work_shape = mask_shape

    theta = theta_filtergrid(work_shape, quadrant_shift=False)
    # Convert from angles in radian range [-pi, +pi] to degree range [0, 360]
    theta = theta / np.pi * 180
    theta[np.where(theta < 0)] += 360

    masks = []
    sector_size = 360 / n_sectors
    for i in range(n_sectors):
        if clockwise:
            theta = np.flip(theta, axis=1)
            sector_start = start_angle - i * sector_size
            sector_end = start_angle - (i + 1) * sector_size
        else:
            sector_start = start_angle + i * sector_size
            sector_end = start_angle + (i + 1) * sector_size

        sector_start = sector_start % 360
        sector_end = sector_end % 360

        mask = np.zeros(work_shape)
        # Handle clockwise and counter-clockwise sector rotation
        if clockwise:

            if sector_start > sector_end:
                # If rotating clockwise the start angle is bigger than the end angle
                selection = np.where(
                    np.logical_and(theta <= sector_start, theta > sector_end))
            else:
                # If this is not the case, only the end angle has crossed the 0°
                selection = np.where(
                    np.logical_or(theta <= sector_start, theta > sector_end))
        else:
            if sector_start < sector_end:
                # If rotating counter-clockwise the start angle is smaller than the end
                selection = np.where(
                    np.logical_and(theta >= sector_start, theta < sector_end))
            else:
                # If this is not the case only the end angle has crossed the 360°
                selection = np.where(
                    np.logical_or(theta >= sector_start, theta < sector_end))

        mask[selection] = 1

        if smooth_edges:
            mask = transform.resize(mask, mask_shape)

        masks.append(mask)

    return masks


# Cache created grids. When quantifying many volumes you need the grid for both laterality and possibly different localizer sizes.
@functools.lru_cache(8, typed=False)
def create_grid_regions(
    mask_shape: tuple[int, int],
    radii: Sequence[int],
    n_sectors: Sequence[int],
    offsets: Sequence[int],
    clockwise: bool,
    smooth_edges: bool = False,
) -> list[npt.NDArray[Any]]:
    """Create sectorized circular region masks.

    First circular masks with the provided radii are generated. Then ring masks
    are created by subtracting the first circular mask from the second and so
    on.
    If you want the complete ring, set the respective n_sectors entry to 1. You  can split
    the ring into n sectors by setting the respective entry to n.
    Setting a number in `offsets` rotates the respective ring sectors by n
    degree.

    Args:
        mask_shape: Output shape of the computed masks
        radii: Ascending radii of the circular regions in pixels
        n_sectors: Number of sectors corresponding to the radii
        offsets: Angular offset of first sector corresponding to the radii
        clockwise: If True sectors are added clockwise starting from the start_angles
        smooth_edges: If True, compute non binary masks where edges might be shared between adjacent regions

    Returns:
    """
    # Create circles
    circles = []
    for radius in radii:
        circles.append(circle_mask(radius, mask_shape, smooth_edges))

    level_sector_parts = []
    for n_sec, start_angle in zip(n_sectors, offsets):
        if n_sec is not None:
            level_sector_parts.append(
                create_sectors(
                    mask_shape,
                    n_sectors=n_sec,
                    start_angle=start_angle,
                    clockwise=clockwise,
                    smooth_edges=smooth_edges,
                ))

    rings = [circles[0]]
    for i, _ in enumerate(circles):
        if i + 1 >= len(circles):
            break
        elif not radii[i] < radii[i + 1]:
            break
        else:
            rings.append(-circles[i] + circles[i + 1])

    pairs = zip(rings, level_sector_parts)

    all_masks = []
    for cir, sectors in pairs:
        for sec in sectors:
            all_masks.append(cir * sec)

    return all_masks


def grid(
    mask_shape: tuple[int, int],
    radii: Union[Sequence[Union[int, float]], int, float],
    laterality: str,
    n_sectors: Union[Sequence[Union[int, float]], int, float] = 1,
    offsets: Union[Sequence[Union[int, float]], int, float] = 0,
    center: Optional[tuple] = None,
    smooth_edges: bool = False,
    radii_scale: Union[int, float] = 1,
) -> dict[str, npt.NDArray[Any]]:
    """Create a quantification grid.

    Args:
        mask_shape: Output shape of the computed masks
        radii: Ascending radii of the circular regions in pixels
        laterality: OD/OS depending for which eye to compute the grid
        n_sectors: Number of sectors corresponding to the radii
        offsets: Sector offsets from the horizonal line on the nasal side in degree
        center: Center location of the computed masks
        smooth_edges: If True, compute non binary masks where edges might be shared between adjacent regions
        radii_scale:

    Returns:
    """
    # Make sure radii, n_sectors and offsets are lists even if you get numbers or tuples
    radii = [radii] if isinstance(radii, (int, float)) else list(radii)
    if not sorted(radii) == radii:
        raise ValueError('radii have to be given in ascending order')
    input_radii = radii
    radii = [r / radii_scale for r in radii]

    n_sectors = [n_sectors] if isinstance(n_sectors,
                                          (int, float)) else list(n_sectors)
    if len(n_sectors) == 1:
        n_sectors = n_sectors * len(radii)

    offsets = [offsets] if isinstance(offsets, (int, float)) else list(offsets)
    if len(offsets) == 1:
        offsets = offsets * len(radii)

    clockwise = False
    masks = create_grid_regions(
        mask_shape,
        tuple(radii),
        tuple(n_sectors),
        tuple(offsets),
        clockwise,
        smooth_edges,
    )

    names = []
    radii = [0.0] + radii
    input_radii = [0] + input_radii
    for i, r in enumerate(radii):
        if i + 1 >= len(radii):
            break
        for s in range(n_sectors[i]):
            names.append(
                f'Radius: {input_radii[i]}-{input_radii[i+1]} Sector: {s}')

    masks = {name: mask for name, mask in zip(names, masks)}
    if laterality == 'OS':
        masks = {name: np.flip(m, axis=1) for name, m in masks.items()}
    elif laterality == 'OD':
        pass
    else:
        raise ValueError('laterality has to be one of OD/OS')

    if center is not None:
        translation = transform.AffineTransform(translation=np.array(center) -
                                                np.array(mask_shape) / 2)
        masks = {
            name: transform.warp(masks[name], translation.inverse)
            for name in masks.keys()
        }

    return masks


@functools.lru_cache(maxsize=8)
def filtergrid(size: Shape,
               quadrant_shift: bool = True,
               normalize: bool = True) -> tuple[np.ndarray, np.ndarray]:
    """Generates grid for constructing frequency domain filters.

    Coordinate matrices for x and y value for a 2D array. The out can be quadrant shifted
    and / or normalized. This is basically a wrapper around np.meshgrid.

    Inspired by filtergrid.m found at https://www.peterkovesi.com/matlabfns/

    Args:
        size: Size of the filter
        quadrant_shift: Quadrant shift such that 0 values / frequencies are at the corners
        normalize: Normalize the range to [-0.5,0.5]

    Returns:
    """
    if type(size) is int:
        rows = cols = size
    else:
        rows = size[0]
        cols = size[1]

    range_1 = np.linspace(-(cols // 2), np.floor((cols - 1) / 2), cols)
    range_2 = np.linspace(-(rows // 2), np.floor((rows - 1) / 2), rows)

    if normalize:
        range_1 = range_1 / cols
        range_2 = range_2 / rows

    x, y = np.meshgrid(range_1, range_2)

    # Quadrant shift so that filters are constructed with 0 frequency at the corners
    if quadrant_shift:
        x = np.fft.ifftshift(x)
        y = np.fft.ifftshift(y)

    return x.T, y.T


@functools.lru_cache(maxsize=8)
def radius_filtergrid(size: Shape,
                      quadrant_shift: bool = True,
                      normalize: bool = True) -> np.ndarray:
    """Radius Filtergrid.

    A matrix containing the radius from the center. This radius is in range [0, 0.5] if normalized.
    The result can be quadrant shifted such that the 0 values are in the corners.

    Args:
        size: Size of the filter
        quadrant_shift: Quadrant shift such that 0 values / frequencies are at the corners
        normalize: Normalize radius to [0 ,0.5]

    Returns:
    """
    x, y = filtergrid(size, quadrant_shift, normalize)
    radius = np.sqrt(x**2 + y**2)
    return radius


@functools.lru_cache(maxsize=8)
def theta_filtergrid(size: Shape, quadrant_shift: bool = True) -> np.ndarray:
    """Theta Filtergrid.

    A matrix containing the polar angle in radian at the respective position for a circle centered in the matrix.
    The result can be returned quadrant shifted. The angle is 0 for all points on the positive x-axis.
    The angles are pi/2 (90°) and -pi/2 (-90°) on the positive and negative y-axis respectively. On the negative
    x-axis the angle is pi (180°). If you need the angle to be in range [0, 2pi] instead of [-pi, pi], you can simply
    add 2pi whenever the angle is negative.

    Args:
        size: Size of the filter
        quadrant_shift: Quadrant shift such that 0 values / frequencies are at the corners

    Returns:
    """
    y, x = filtergrid(size, quadrant_shift)

    # Matrix values contain polar angle.
    # 0 angle starts on the horizontal line and runs counter clock-wise
    theta = np.arctan2(-y, x)

    return theta



from __future__ import annotations

from typing import Any, Optional, TYPE_CHECKING, Union

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from numpy import typing as npt
import numpy as np
from scipy import ndimage
from skimage import transform

from eyepy.core.annotations import EyeEnfaceFoveaAnnotation
from eyepy.core.annotations import EyeEnfaceOpticDiscAnnotation
from eyepy.core.annotations import EyeEnfacePixelAnnotation
from eyepy.core.eyemeta import EyeEnfaceMeta
from eyepy.core.plotting import plot_scalebar
from eyepy.core.plotting import plot_watermark

if TYPE_CHECKING:
    pass


class EyeEnface:
    """Enface (2D projection) image with optional anatomical annotations.

    **Coordinate System Convention:**
    EyeEnface uses (row, col) image pixel coordinates:
    - row: vertical axis, increases downward (corresponds to y)
    - col: horizontal axis, increases rightward (corresponds to x)

    All image operations (scaling, translation, rotation) work in this space.
    Physical coordinates (x, y) in millimeters are converted to (col, row)
    format when needed for internal calculations.

    Annotations (optic disc, fovea, pixel annotations) are stored and
    transformed in (row, col) coordinates.
    """

    def __init__(self, data: npt.NDArray[np.int64],
                 meta: EyeEnfaceMeta,
                 optic_disc: Optional[EyeEnfaceOpticDiscAnnotation] = None,
                 fovea: Optional[EyeEnfaceFoveaAnnotation] = None) -> None:
        """Initialize EyeEnface.

        Args:
            data: Enface image data
            meta: Metadata for the enface image
            optic_disc: Optional EyeEnfaceOpticDiscAnnotation
            fovea: Optional EyeEnfaceFoveaAnnotation
        """
        self.data = data
        self._area_maps = []
        self.meta = meta
        self._optic_disc = optic_disc
        self._fovea = fovea

        # Validate and set laterality if both optic disc and fovea are provided
        if optic_disc is not None and fovea is not None:
            self._infer_and_validate_laterality()

    @property
    def optic_disc(self) -> Optional[EyeEnfaceOpticDiscAnnotation]:
        """Get the optic disc annotation.

        Returns:
            The optic disc annotation or None if not set
        """
        return self._optic_disc

    @optic_disc.setter
    def optic_disc(self, value: Optional[EyeEnfaceOpticDiscAnnotation]) -> None:
        """Set the optic disc annotation.

        If both optic_disc and fovea are set, laterality will be inferred (if not already set)
        and validated.

        Args:
            value: The optic disc annotation or None

        Raises:
            ValueError: If laterality validation fails
        """
        self._optic_disc = value
        # Validate and infer laterality if both annotations are now present
        if self._optic_disc is not None and self._fovea is not None:
            self._infer_and_validate_laterality()

    @property
    def fovea(self) -> Optional[EyeEnfaceFoveaAnnotation]:
        """Get the fovea annotation.

        Returns:
            The fovea annotation or None if not set
        """
        return self._fovea

    @fovea.setter
    def fovea(self, value: Optional[EyeEnfaceFoveaAnnotation]) -> None:
        """Set the fovea annotation.

        If both optic_disc and fovea are set, laterality will be inferred (if not already set)
        and validated.

        Args:
            value: The fovea annotation or None

        Raises:
            ValueError: If laterality validation fails
        """
        self._fovea = value
        # Validate and infer laterality if both annotations are now present
        if self._optic_disc is not None and self._fovea is not None:
            self._infer_and_validate_laterality()

    def _infer_and_validate_laterality(self) -> None:
        """Infer laterality if not set, then validate it.

        If laterality is not set in metadata, infer it from the relative positions
        of optic disc and fovea and set it in the metadata. Then validate that the
        laterality matches the anatomical positions.

        Raises:
            ValueError: If laterality validation fails
        """
        if self._optic_disc is None or self._fovea is None:
            return

        laterality = self.meta.get('laterality', None)

        # If laterality is not set, infer and set it
        if laterality is None:
            inferred = self._infer_laterality()
            if inferred is not None:
                self.meta['laterality'] = inferred
                return

        # Now validate the laterality
        self._validate_laterality()

    def _infer_laterality(self) -> Optional[str]:
        """Infer laterality from the relative positions of optic disc and fovea.

        For anatomically correct images:
        - If optic disc is to the right of fovea -> Right eye (OD)
        - If optic disc is to the left of fovea -> Left eye (OS)

        Returns:
            'OD' for right eye, 'OS' for left eye, or None if cannot infer
        """
        if self._optic_disc is None or self._fovea is None:
            return None

        od_center = self._optic_disc.center
        fovea_center = self._fovea.center

        # Compare col-coordinates (center returns (row, col) format)
        od_col = od_center[1]
        fovea_col = fovea_center[1]

        # If optic disc is to the right of fovea -> right eye
        if od_col > fovea_col:
            return 'OD'
        # If optic disc is to the left of fovea -> left eye
        elif od_col < fovea_col:
            return 'OS'
        else:
            # They are at the same column position, cannot infer
            return None

    def _validate_laterality(self) -> None:
        """Validate that laterality matches the relative position of optic disc
        and fovea.

        For a right eye (OD), the optic disc should be to the right of the fovea (higher col).
        For a left eye (OS), the optic disc should be to the left of the fovea (lower col).

        Raises:
            ValueError: If the laterality doesn't match the anatomical positions
        """
        if self._optic_disc is None or self._fovea is None:
            return

        laterality = self.meta.get('laterality', None)
        if laterality is None:
            # No laterality info, skip validation
            return

        # Get centers
        od_center = self._optic_disc.center
        fovea_center = self._fovea.center

        # Compare col-coordinates (center returns (row, col) format)
        od_col = od_center[1]
        fovea_col = fovea_center[1]

        # For right eye (OD): optic disc should be to the right of fovea
        # For left eye (OS): optic disc should be to the left of fovea
        if laterality.upper() in ['OD', 'R', 'RIGHT']:
            if od_col <= fovea_col:
                raise ValueError(
                    f'Laterality mismatch: Right eye (OD) expects optic disc to the right of fovea, '
                    f'but optic disc is at col={od_col:.1f} and fovea is at col={fovea_col:.1f}'
                )
        elif laterality.upper() in ['OS', 'L', 'LEFT']:
            if od_col >= fovea_col:
                raise ValueError(
                    f'Laterality mismatch: Left eye (OS) expects optic disc to the left of fovea, '
                    f'but optic disc is at col={od_col:.1f} and fovea is at col={fovea_col:.1f}'
                )

    @property
    def area_maps(self) -> dict[str, EyeEnfacePixelAnnotation]:
        """

        Returns:

        """
        # Create a dict to access area_maps by their name
        return {am.name: am for am in self._area_maps}

    def add_area_annotation(self,
                            area_map: Optional[npt.NDArray[np.bool_]] = None,
                            meta: Optional[dict] = None,
                            **kwargs: Any) -> EyeEnfacePixelAnnotation:
        """

        Args:
            area_map:
            meta:
            **kwargs:

        Returns:

        """
        if meta is None:
            meta = {}
        meta.update(**kwargs)
        area_annotation = EyeEnfacePixelAnnotation(self, area_map, meta)
        self._area_maps.append(area_annotation)
        return area_annotation

    @property
    def scale_x(self) -> float:
        """

        Returns:

        """
        return self.meta['scale_x']

    @scale_x.setter
    def scale_x(self, value: float) -> None:
        self.meta['scale_x'] = value

    @property
    def scale_y(self) -> float:
        """

        Returns:

        """
        return self.meta['scale_y']

    @scale_y.setter
    def scale_y(self, value: float) -> None:
        self.meta['scale_y'] = value

    @property
    def scale_unit(self) -> str:
        """

        Returns:

        """
        return self.meta['scale_unit']

    @scale_unit.setter
    def scale_unit(self, value: str) -> None:
        self.meta['scale_unit'] = value

    @property
    def size_x(self) -> int:
        """

        Returns:

        """
        return self.shape[1]

    @property
    def size_y(self) -> int:
        """

        Returns:

        """
        return self.shape[0]

    @property
    def laterality(self) -> str:
        """

        Returns:

        """
        return self.meta['laterality']

    @property
    def shape(self) -> tuple[int, int]:
        """

        Returns:

        """
        return self.data.shape

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        region: tuple[slice, slice] = np.s_[:, :],
        scalebar: Union[bool, str] = 'botleft',
        scalebar_kwargs: Optional[dict[str, Any]] = None,
        watermark: bool = True,
        areas: Union[bool, list[str]] = True,
        plot_optic_disc: bool = True,
        plot_fovea: bool = True,
        area_kwargs: Optional[dict[str, Any]] = None,
        optic_disc_kwargs: Optional[dict[str, Any]] = None,
        fovea_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        """

        Args:
            ax: Axes to plot on. If not provided plot on the current axes (plt.gca()).
            region: Region of the localizer to plot (default: `np.s_[:, :]`)
            scalebar: Position of the scalebar, one of "topright", "topleft", "botright", "botleft" or `False` (default: "botleft"). If `True` the scalebar is placed in the bottom left corner. You can custumize the scalebar using the `scalebar_kwargs` argument.
            scalebar_kwargs: Optional keyword arguments for customizing the scalebar. Check the documentation of [plot_scalebar][eyepy.core.plotting.plot_scalebar] for more information.
            watermark: If `True` plot a watermark on the image (default: `True`). When removing the watermark, please consider to cite eyepy in your publication.
            areas: If `True` plot all area annotations (default: `True`). If a list of strings is given, plot the area annotations with the given names.
            plot_optic_disc: If `True` and optic_disc is available, plot the optic disc annotation (default: `True`).
            plot_fovea: If `True` and fovea is available, plot the fovea annotation (default: `True`).
            area_kwargs: Optional keyword arguments for customizing the area annotations (alpha, etc.).
            optic_disc_kwargs: Optional keyword arguments for customizing the optic disc plot (color, linewidth, etc.).
            fovea_kwargs: Optional keyword arguments for customizing the fovea plot (color, marker, markersize, etc.).
        Returns:
            None

        """
        ax = plt.gca() if ax is None else ax
        vmin = np.min(self.data)
        vmax = np.max(self.data)

        ax.imshow(self.data[region], cmap='gray', vmin=vmin, vmax=vmax)

        # Make sure tick labels match the image region
        y_start = region[0].start if region[0].start is not None else 0
        x_start = region[1].start if region[1].start is not None else 0
        y_end = region[0].stop if region[0].stop is not None else self.size_y
        x_end = region[1].stop if region[1].stop is not None else self.size_x

        # Ticks are not clipped to the image region. Clip them here.
        yticks = ax.get_yticks()
        yticks = yticks[np.nonzero(
            np.logical_and(yticks >= 0, yticks <= y_end - y_start - 1))]
        xticks = ax.get_xticks()
        xticks = xticks[np.nonzero(
            np.logical_and(xticks >= 0, xticks <= x_end - x_start - 1))]

        # Set clipped ticks (this is only necessary because we change the labels later)
        ax.set_yticks(yticks)
        ax.set_xticks(xticks)

        # Set labels to ticks + start of the region as an offset
        ax.set_yticklabels([str(int(t + y_start)) for t in yticks])
        ax.set_xticklabels([str(int(t + x_start)) for t in xticks])

        if scalebar:
            if scalebar_kwargs is None:
                scalebar_kwargs = {}

            scale_unit = self.meta['scale_unit']
            scalebar_kwargs = {
                **{
                    'scale': (self.scale_x, self.scale_y),
                    'scale_unit': scale_unit
                },
                **scalebar_kwargs
            }

            if not 'pos' in scalebar_kwargs:
                sx = x_end - x_start
                sy = y_end - y_start

                if scalebar is True:
                    scalebar = 'botleft'

                if scalebar == 'botleft':
                    scalebar_kwargs['pos'] = (sx - 0.95 * sx, 0.95 * sy)
                elif scalebar == 'botright':
                    scalebar_kwargs['pos'] = (0.95 * sx, 0.95 * sy)
                    scalebar_kwargs['flip_x'] = True
                elif scalebar == 'topleft':
                    scalebar_kwargs['pos'] = (sx - 0.95 * sx, 0.05 * sy)
                    scalebar_kwargs['flip_y'] = True
                elif scalebar == 'topright':
                    scalebar_kwargs['pos'] = (0.95 * sx, 0.05 * sy)
                    scalebar_kwargs['flip_x'] = True
                    scalebar_kwargs['flip_y'] = True

            plot_scalebar(ax=ax, **scalebar_kwargs)

        if watermark:
            plot_watermark(ax)

        # Handle areas parameter
        if not areas:
            areas_to_plot = []
        elif areas is True:
            areas_to_plot = list(self.area_maps.keys())
        else:
            areas_to_plot = areas

        # Set default area_kwargs if not provided
        if area_kwargs is None:
            area_kwargs = {'alpha': 0.5}

        # Plot area annotations
        for area_name in areas_to_plot:
            if area_name not in self.area_maps:
                continue

            area_annotation = self.area_maps[area_name]
            data = area_annotation.data[region]
            visible = np.zeros(data.shape, dtype=bool)
            visible[data != 0] = 1.0

            # Get color from metadata or use default
            meta = area_annotation.meta
            color = meta.get('color', 'red')
            color = mcolors.to_rgba(color)

            # Create a 0 radius circle patch as dummy for the area label
            patch = mpatches.Circle((0, 0), radius=0, color=color, label=area_name)
            ax.add_patch(patch)

            # Create plot_data by tiling the color vector over the plotting shape
            plot_data = np.tile(np.array(color), data.shape + (1, ))
            # Now turn the alpha channel 0 where the mask is 0 and adjust the remaining alpha
            plot_data[..., 3] *= visible * area_kwargs.get('alpha', 0.5)

            ax.imshow(
                plot_data,
                interpolation='none',
            )

        # Plot optic disc if available
        if plot_optic_disc and self.optic_disc is not None:
            if optic_disc_kwargs is None:
                optic_disc_kwargs = {}

            # Calculate region offset
            y_start = region[0].start if region[0].start is not None else 0
            x_start = region[1].start if region[1].start is not None else 0

            # Delegate to the optic disc's plot method
            self.optic_disc.plot(ax=ax, offset=(x_start, y_start), **optic_disc_kwargs)

        # Plot fovea if available
        if plot_fovea and self.fovea is not None:
            if fovea_kwargs is None:
                fovea_kwargs = {}

            # Calculate region offset
            y_start = region[0].start if region[0].start is not None else 0
            x_start = region[1].start if region[1].start is not None else 0

            # Delegate to the fovea's plot method
            self.fovea.plot(ax=ax, offset=(x_start, y_start), **fovea_kwargs)

    def scale(self, scale_y: float, scale_x: float,
              order: int = 1, mode: str = 'constant', cval: float = 0.0) -> 'EyeEnface':
        """Scale the enface image and all annotations.

        Returns a new EyeEnface instance with scaled image data and transformed annotations.
        The original EyeEnface remains unchanged.

        Args:
            scale_y: Scaling factor for the y-axis
            scale_x: Scaling factor for the x-axis
            order: The order of interpolation (0=nearest, 1=bilinear, 3=bicubic, default: 1)
            mode: How to handle values outside the boundaries ('constant', 'nearest', 'reflect', 'wrap')
            cval: Value used for points outside the boundaries if mode='constant'

        Returns:
            New EyeEnface instance with scaled data and annotations
        """
        # Create affine transformation matrix for scaling
        matrix = np.array([
            [scale_y, 0, 0],
            [0, scale_x, 0],
            [0, 0, 1]
        ])

        # Calculate output shape
        output_shape = (int(self.shape[0] * scale_y), int(self.shape[1] * scale_x))

        return self.transform(matrix, output_shape=output_shape, order=order, mode=mode, cval=cval)

    def translate(self, drow: float, dcol: float,
                  order: int = 1, mode: str = 'constant', cval: float = 0.0) -> 'EyeEnface':
        """Translate the enface image and all annotations.

        Returns a new EyeEnface instance with translated image data and transformed annotations.
        The original EyeEnface remains unchanged.

        Args:
            drow: Translation offset in the row direction (vertical, pixels)
            dcol: Translation offset in the col direction (horizontal, pixels)
            order: The order of interpolation (0=nearest, 1=bilinear, 3=bicubic, default: 1)
            mode: How to handle values outside the boundaries ('constant', 'nearest', 'reflect', 'wrap')
            cval: Value used for points outside the boundaries if mode='constant'

        Returns:
            New EyeEnface instance with translated data and annotations
        """
        # Create affine transformation matrix for translation
        # Matrix format: [[1, 0, drow], [0, 1, dcol], [0, 0, 1]]
        matrix = np.array([
            [1, 0, drow],
            [0, 1, dcol],
            [0, 0, 1]
        ])

        return self.transform(matrix, order=order, mode=mode, cval=cval)

    def rotate(self, angle: float, center: Optional[tuple[float, float]] = None,
               order: int = 1, mode: str = 'constant', cval: float = 0.0) -> 'EyeEnface':
        """Rotate the enface image and all annotations.

        Returns a new EyeEnface instance with rotated image data and transformed annotations.
        The original EyeEnface remains unchanged.

        Args:
            angle: Rotation angle in degrees (positive = counter-clockwise)
            center: Center of rotation (row, col). If None, uses image center
            order: The order of interpolation (0=nearest, 1=bilinear, 3=bicubic, default: 1)
            mode: How to handle values outside the boundaries ('constant', 'nearest', 'reflect', 'wrap')
            cval: Value used for points outside the boundaries if mode='constant'

        Returns:
            New EyeEnface instance with rotated data and annotations
        """
        if center is None:
            center = (self.shape[0] / 2, self.shape[1] / 2)

        # Build rotation matrix in (row, col) coordinates for counter-clockwise rotation
        # Standard mathematical rotation: positive angle = counter-clockwise
        # The standard counter-clockwise rotation matrix is:
        # [[cos(θ), -sin(θ)], [sin(θ), cos(θ)]]
        # But with (row, col) format where row increases downward (y-axis flipped),
        # we need: [[cos(θ), sin(θ)], [-sin(θ), cos(θ)]] to get counter-clockwise rotation
        center_row, center_col = center
        angle_rad = np.deg2rad(angle)
        cos_a = np.cos(angle_rad)
        sin_a = np.sin(angle_rad)

        # Rotation matrix for (row, col) format with counter-clockwise rotation around (center_row, center_col)
        matrix_rowcol = np.array([
            [cos_a, -sin_a, center_row - cos_a * center_row + sin_a * center_col],
            [sin_a, cos_a, center_col - sin_a * center_row - cos_a * center_col],
            [0, 0, 1]
        ])

        return self.transform(matrix_rowcol, order=order, mode=mode, cval=cval)

    def standard_rotation(self, order: int = 1, mode: str = 'constant', cval: float = 0.0) -> 'EyeEnface':
        """Rotate the enface to align optic disc and fovea centers horizontally.

        This method rotates the image so that the optic disc and fovea centers lie on the
        same horizontal line. The rotation is performed around the midpoint between the OD
        and fovea, preserving their relative positions. The rotation direction depends on
        the laterality:
        - For right eyes (OD): optic disc on the right, fovea on the left
        - For left eyes (OS): optic disc on the left, fovea on the right

        This creates a standardized orientation useful for comparison across different images.

        Requires both optic_disc and fovea annotations to be present.

        Args:
            order: The order of interpolation (0=nearest, 1=bilinear, 3=bicubic, default: 1)
            mode: How to handle values outside the boundaries ('constant', 'nearest', 'reflect', 'wrap')
            cval: Value used for points outside the boundaries if mode='constant'

        Returns:
            New EyeEnface instance with standardized rotation

        Raises:
            ValueError: If optic_disc or fovea annotations are missing
        """
        if self._optic_disc is None:
            raise ValueError('standard_rotation requires optic_disc annotation')
        if self._fovea is None:
            raise ValueError('standard_rotation requires fovea annotation')

        # Get centers of optic disc and fovea
        od_center = self.optic_disc.center
        fovea_center = self.fovea.center

        # Determine laterality from metadata or infer from positions
        laterality = self.meta.get('laterality', None)

        # Calculate vector from OD to fovea (in row, col coordinates)
        # center returns (row, col) tuples
        d_row = fovea_center[0] - od_center[0]
        d_col = fovea_center[1] - od_center[1]

        # Calculate the current angle of the OD-to-fovea line
        # In (row, col) space, horizontal line has d_row=0, so angle from horizontal is arctan2(d_row, d_col)
        # Positive angle means counter-clockwise from positive column direction (pointing right)
        current_angle_rad = np.arctan2(d_row, d_col)

        # Determine target angle based on laterality (in row, col coordinates)
        # For OD (right eye): OD on right, fovea on left -> line points left (180°)
        # For OS (left eye): OD on left, fovea on right -> line points right (0°)
        if laterality is not None:
            laterality_upper = str(laterality).upper()
            if laterality_upper in ['OD', 'R', 'RIGHT']:
                target_angle_rad = np.pi  # Pointing left
            elif laterality_upper in ['OS', 'L', 'LEFT']:
                target_angle_rad = 0.0  # Pointing right
            else:
                # Unknown laterality, default to horizontal alignment with fovea to the right
                target_angle_rad = 0.0
        else:
            # No laterality info, align horizontally with fovea to the right
            target_angle_rad = 0.0

        # Calculate rotation angle needed
        rotation_angle_rad = target_angle_rad - current_angle_rad

        # Normalize to [-π, π] to take the shortest rotation path
        rotation_angle_rad = np.arctan2(np.sin(rotation_angle_rad), np.cos(rotation_angle_rad))
        rotation_angle_deg = np.rad2deg(rotation_angle_rad)

        # Rotate around the midpoint between OD and fovea to keep them centered
        # This preserves their relative position better than rotating around image center
        midpoint_row = (od_center[0] + fovea_center[0]) / 2
        midpoint_col = (od_center[1] + fovea_center[1]) / 2
        center = (midpoint_row, midpoint_col)

        return self.rotate(-rotation_angle_deg, center=center, order=order, mode=mode, cval=cval)

    def transform(self, matrix: npt.NDArray[np.float64],
                  output_shape: Optional[tuple[int, int]] = None,
                  order: int = 1, mode: str = 'constant', cval: float = 0.0) -> 'EyeEnface':
        """Apply an affine transformation to the enface image and all annotations.

        Returns a new EyeEnface instance with transformed image data and annotations.
        The original EyeEnface remains unchanged. The metadata is updated to reflect
        any scaling in the transformation.

        Args:
            matrix: 3x3 affine transformation matrix in homogeneous coordinates or 2x3 matrix
            output_shape: Shape of the output image (height, width). If None, uses input shape
            order: The order of interpolation (0=nearest, 1=bilinear, 3=bicubic, default: 1)
            mode: How to handle values outside the boundaries ('constant', 'nearest', 'reflect', 'wrap')
            cval: Value used for points outside the boundaries if mode='constant'

        Returns:
            New EyeEnface instance with transformed data and annotations
        """
        if output_shape is None:
            output_shape = self.shape

        # Convert 3x3 to 2x3 for annotations if needed
        if matrix.shape == (3, 3):
            # Matrix is in (row, col) format - use directly for annotations
            annotation_matrix = matrix[:2, :]
        elif matrix.shape == (2, 3):
            annotation_matrix = matrix
        else:
            raise ValueError('Matrix must be 3x3 or 2x3')

        # For image transformation with skimage, we need to convert to (x, y) format
        # skimage's AffineTransform works with (x, y) format, not (row, col)
        # So we need to swap the matrix dimensions: swap rows and columns
        if matrix.shape == (2, 3):
            # Convert 2x3 to 3x3
            full_matrix = np.vstack([matrix, [0, 0, 1]])
        else:
            full_matrix = matrix

        # Convert from (row, col) to (x, y) by swapping the matrix rows/cols
        # If M transforms (row, col) -> (row', col'), then for (x, y) = (col, row):
        # [row']   [a  b  tx]   [row]       [col']   [b  a  ty]   [col]
        # [col'] = [c  d  ty] * [col]  =>   [row'] = [d  c  tx] * [row]
        skimage_matrix = np.array([
            [full_matrix[1, 1], full_matrix[1, 0], full_matrix[1, 2]],
            [full_matrix[0, 1], full_matrix[0, 0], full_matrix[0, 2]],
            [0, 0, 1]
        ])

        # Extract scaling factors from the transformation matrix
        # The scaling factors are the norms of the first two columns
        scale_y = np.linalg.norm(annotation_matrix[:, 0])
        scale_x = np.linalg.norm(annotation_matrix[:, 1])

        # Create updated metadata with adjusted scale_x and scale_y
        # When the image is scaled up, each pixel represents less physical distance
        # Only update metadata if it's a real EyeEnfaceMeta object, not a mock
        if isinstance(self.meta, EyeEnfaceMeta):
            # Real EyeEnfaceMeta object - use the copy() method
            updated_meta = self.meta.copy()
            if 'scale_x' in updated_meta:
                updated_meta['scale_x'] = updated_meta['scale_x'] / scale_x
            if 'scale_y' in updated_meta:
                updated_meta['scale_y'] = updated_meta['scale_y'] / scale_y
        else:
            # Mock or other type - just reuse it
            updated_meta = self.meta

        # Create AffineTransform from matrix
        # Note: scikit-image uses inverse matrix convention
        tform = transform.AffineTransform(matrix=skimage_matrix)

        # Transform the image data
        transformed_data = transform.warp(
            self.data,
            tform.inverse,
            output_shape=output_shape,
            order=order,
            mode=mode,
            cval=cval,
            preserve_range=True
        ).astype(self.data.dtype)

        # Transform annotations
        transformed_optic_disc = self._optic_disc.transform(annotation_matrix) if self._optic_disc is not None else None
        transformed_fovea = self._fovea.transform(annotation_matrix) if self._fovea is not None else None

        # Create new EyeEnface instance with updated metadata
        new_enface = EyeEnface(
            data=transformed_data,
            meta=updated_meta,
            optic_disc=transformed_optic_disc,
            fovea=transformed_fovea
        )

        # Transform area maps
        for area_map in self._area_maps:
            if area_map.data is not None:
                transformed_area_data = transform.warp(
                    area_map.data.astype(float),
                    tform.inverse,
                    output_shape=output_shape,
                    order=0,  # Use nearest neighbor for binary masks
                    mode=mode,
                    cval=cval,
                    preserve_range=True
                ).astype(bool)
            else:
                transformed_area_data = None

            new_area_map = EyeEnfacePixelAnnotation(
                new_enface,
                transformed_area_data,
                area_map.meta.copy()
            )
            new_enface._area_maps.append(new_area_map)

        return new_enface



from collections import defaultdict

# Plotting config

# Line Style for Layers in B-Scan
layer_kwargs = {'linewidth': 1, 'linestyle': '-'}
slab_kwargs = {'alpha': 0.5}
area_kwargs = {'alpha': 0.5}
ascan_kwargs = {'alpha': 0.5}

# Line Style for B-Scan positions on Slo
line_kwargs = {'linewidth': 0.2, 'linestyle': '-', 'color': 'limegreen'}

# Colors for different Layers
_layer_colors = {
    'BM': 'F77189',
    'RPE': '97A431',
    'PR1': '36ADA4',
    'EZ': '36ADA4',
    'PR2': 'A48CF4',
    'IZ': 'A48CF4',
    'ELM': 'E68332',
    'ONL': '50B131',
    'OPL': '38AABF',
    'INL': 'E866F4',
    'IPL': 'BB9832',
    'GCL': '34AF84',
    'RNFL': '3BA3EC',
    'NFL': '3BA3EC',
    'ILM': 'F668C2',
}

layer_colors = defaultdict(lambda: 'FF0000', **_layer_colors)

# Colors for different Slabs
_slab_colors = {
    'NFLVP': '00598C',
    'SVP': '0057E5',
    'ICP': '41B6E6',
    'DCP': '3EA908',
    'SVC': 'F0CC2E',
    'DVC': 'EC894D',
    'AVAC': 'E50000',
    'RET': 'E641B6',
}

slab_colors = defaultdict(lambda: 'FBB2C4', **_slab_colors)

# Colors for different Area annotations
_area_colors = {
}

area_colors = defaultdict(lambda: 'FF0000', **_area_colors)



import logging

import numpy as np
import numpy.typing as npt
from scipy import ndimage as ndimage

logger = logging.getLogger(__name__)


def filter_by_depth(drusen_map: npt.NDArray[np.bool_],
                    minimum_depth: int = 2) -> npt.NDArray[np.bool_]:
    """

    Args:
        drusen_map:
        minimum_depth:

    Returns:

    """
    filtered_drusen = np.copy(drusen_map)
    if minimum_depth == 0:
        return drusen_map
    # get array where connected components get same label
    connected_component_array, _ = ndimage.label(drusen_map, output=None)
    # Go through each component, sum it along axis 0 and check max depth against threshold
    max_depths = np.zeros_like(connected_component_array)
    for label, drusen_pos in enumerate(
            ndimage.find_objects(connected_component_array)):
        component_sub_vol = connected_component_array[drusen_pos]
        component_max_depth = np.max(
            np.sum(component_sub_vol == label + 1, axis=0))
        component_sub_vol[component_sub_vol == label + 1] = component_max_depth
        max_depths[drusen_pos] = component_sub_vol
    filtered_drusen[max_depths < minimum_depth] = False
    return filtered_drusen.astype(bool)


def filter_by_height_enface(drusen_map: npt.NDArray[np.bool_],
                            minimum_height: int = 2) -> npt.NDArray[np.bool_]:
    """

    Args:
        drusen_map:
        minimum_height:

    Returns:

    """
    if minimum_height == 0:
        return drusen_map

    projection = np.sum(drusen_map, axis=1,
                        keepdims=True)  # Shape (n_bscans, width)

    # Find connected components in the enface projection
    connected_component_array, num_drusen = ndimage.label(projection != 0)
    # print(connected_component_array.shape)
    max_heights = np.zeros_like(connected_component_array)
    for drusen_pos in ndimage.find_objects(connected_component_array):
        # Work on subvolume for faster processing
        component_sub_vol = connected_component_array[drusen_pos]
        # Find current label (most frequent label in the subvolume)
        label = np.bincount(component_sub_vol[component_sub_vol != 0]).argmax()
        component_max_height = np.max(
            projection[drusen_pos][component_sub_vol == label])
        # Set drusen region to drusen max height
        max_heights[drusen_pos][component_sub_vol ==
                                label] = component_max_height

    filtered_drusen = np.copy(drusen_map)
    indices = np.nonzero(max_heights < minimum_height)
    filtered_drusen[indices[0], :, indices[2]] = False
    return filtered_drusen.astype(bool)



"""Inspired by:

https://github.com/ayl/heyexReader/blob/master/heyexReader/volReader.py
https://github.com/FabianRathke/octSegmentation/blob/master/collector/HDEVolImporter.m
"""
from __future__ import annotations

import datetime
import logging
from typing import Literal
import warnings

import construct as cs
import numpy as np
import skimage
from skimage import transform

from eyepy import __version__
from eyepy import EyeBscanMeta
from eyepy import EyeEnface
from eyepy import EyeEnfaceMeta
from eyepy import EyeVolume
from eyepy import EyeVolumeMeta
from eyepy.core.eyebscan import EyeBscan
from eyepy.core.utils import to_vol_intensity
from eyepy.io.utils import _compute_localizer_oct_transform
from eyepy.io.utils import Bscan
from eyepy.io.utils import FloatDate
from eyepy.io.utils import IntDate
from eyepy.io.utils import Localizer
from eyepy.io.utils import Segmentations

logger = logging.getLogger('eyepy.io.HeVolReader')

bscan_format = cs.Struct(
    'version' /
    cs.Default(cs.PaddedString(12, 'ascii'), f'ep-v{__version__}'.ljust(12)) *
    'Version identifier',
    'bscan_hdr_size' / cs.Int32ul *
    'Size of the B-Scan header in bytes. It is identical to the same value of the file header.',
    'start_x' / cs.Float64l *
    "X-Coordinate of the B-Scan's start point in mm.",
    'start_y' / cs.Float64l *
    "Y-Coordinate of the B-Scan's start point in mm.",
    'end_x' / cs.Float64l *
    "X-Coordinate of the B-Scan's end point in mm. For circle scans, this is the X-Coordinate of the circle's center point.",
    'end_y' / cs.Float64l *
    "Y-Coordinate of the B-Scan's end point in mm. For circle scans, this is the Y-Coordinate of the circle's center point.",
    'num_seg' / cs.Int32ul * 'Number of segmentation vectors',
    'seg_offset' / cs.Default(cs.Int32ul, 256) *
    'Offset of the array of segmentation vectors relative to the beginning of this B-Scan header.',
    'quality' / cs.Float32l *
    'Image quality measure. If this value does not exist, its value is set to INVALID.',
    'shift' / cs.Int32ul *
    'Horizontal shift (in # of A-Scans) of the classification band against the segmentation lines (for circular scan only).',
    'iv_transformation' / cs.Float32l[6] *
    'Intra volume transformation matrix. The values are only available for volume and radial scans and if alignment is turned off, otherwise the values are initialized to 0.',
    '__empty' / cs.Padding(168) * 'Spare bytes for future use.',
    # cs.Probe(cs.this.bscan_hdr.sizeof(cs.this)),
    'layer_segmentations' / Segmentations *
    'Layer segmentations for num_seg Layers. If layers are not segmented the data is empty',
    '__empty' / cs.Padding(cs.this.bscan_hdr_size - 256 -
                           cs.this.num_seg * 4 * cs.this._.size_x) *
    "B-scan data starts after 'bscan_hdr_size' bytes from the start of the"
    'B-scan. This header contains 256 bytes for general information, then "num_seg" '
    'layer segmentations each composed size_x * 4 bytes. Then we have padding'
    'to fill the bscan_hdr_size.',
    'data' / Bscan * 'B-scan data',
)

vol_format = cs.Struct(
    'version' /
    cs.Default(cs.PaddedString(12, 'ascii'), f'ep-v{__version__}'.ljust(12)) *
    'Version identifier: HSF-OCT-xxx',
    'size_x' / cs.Int32ul *
    'Number of A-Scans in each B-Scan, i.e. B-Scan width in pixel',
    'n_bscans' / cs.Int32ul * 'Number of B-Scans in OCT scan',
    'size_y' / cs.Int32ul *
    'Number of samples in an A-Scan, i.e. B-Scan height in pixel',
    'scale_x' / cs.Float64l * 'Width of a B-Scan pixel in mm',
    'distance' / cs.Float64l * 'Distance between two adjacent B-Scans in mm',
    'scale_y' / cs.Float64l * 'Height of a B-Scan pixel in mm',
    'size_x_slo' / cs.Int32ul * 'Width of the SLO image in pixel',
    'size_y_slo' / cs.Int32ul * 'Height of the SLO image in pixel',
    'scale_x_slo' / cs.Float64l * 'Width of a pixel in the SLO image in mm',
    'scale_y_slo' / cs.Float64l * 'Height of a pixel in the SLO image in mm',
    'field_size_slo' / cs.Default(cs.Int32ul, 0) *
    'Horizontal field size of the SLO image in dgr',
    'scan_focus' / cs.Default(cs.Float64l, 0) * 'Scan focus in dpt',
    'scan_position' / cs.Default(cs.PaddedString(4, 'ascii'), ''.ljust(4)) *
    "Examined eye (zero terminated string). 'OS': Left eye; 'OD': Right eye",
    'exam_time' /
    cs.Default(IntDate, datetime.datetime(year=2000, month=1, day=1)) *
    'Examination time. The structure holds an unsigned 64-bit date and time value. '
    'It is encoded as 100ns units since beginning of January 1, 1601',
    'scan_pattern' / cs.Default(cs.Int32sl, 0) * """Scan pattern type:
    0 = Unknown pattern,
    1 = Single line scan (one B-Scan only),
    2 = Circular scan (one B-Scan only),
    3 = Volume scan in ART mode,
    4 = Fast volume scan,
    5 = Radial scan (aka. star pattern)""",
    'bscan_hdr_size' / cs.Default(cs.Int32ul, 61440) *
    'Size of the Header preceding each B-Scan in bytes',
    'id' / cs.PaddedString(16, 'ascii') *
    'Unique identifier of this OCT-scan (zero terminated string). This is'
    'identical to the number <SerID> that is part of the file name.'
    'Format: n[.m] n and m are numbers. The extension .m exists only for '
    'ScanPattern 1 and 2. Examples: 2390, 3433.2',
    'reference_id' / cs.PaddedString(16, 'ascii') *
    'Unique identifier of the reference OCT-scan (zero terminated string).'
    'Format: see ID, This ID is only present if the OCT-scan is part of a '
    'progression otherwise this string is empty. For the reference scan of '
    'a progression ID and ReferenceID are identical.',
    'pid' / cs.Default(cs.Int32ul, 0) * 'Internal patient ID used by HEYEX.',
    'patient_id' /
    cs.Default(cs.PaddedString(24, 'ascii'), 'unknown'.ljust(24)) *
    'User-defined patient ID (zero terminated string).',
    'dob' / FloatDate * "Patient's date of birth",
    'vid' / cs.Default(cs.Int32ul, 0) * 'Internal visit ID used by HEYEX.',
    'visit_id' /
    cs.Default(cs.PaddedString(24, 'ascii'), 'unknown'.ljust(24)) *
    'User-defined visit ID (zero terminated string). This ID can be defined '
    'in the Comment-field of the Diagnosis-tab of the Examination Data '
    'dialog box. The VisitID must be defined in the first row of the '
    "comment field. It has to begin with an '#' and ends with any "
    'white-space character. It can contain up to 23 alphanumeric '
    "characters (excluding the '#')",
    'visit_date' / FloatDate *
    'Date the visit took place. Identical to the date of an examination tab in HEYEX.',
    'grid_type' / cs.Int32ul *
    'Type of grid used to derive thickness data. 0 No thickness data available, >0 Type'
    ' of grid used to derive thickness values. Thickness data is only available for '
    'ScanPattern 3 and 4.',
    'grid_offset' / cs.Int32ul *
    'File offset of the thickness data in the file. If GridType is 0, GridOffset is 0.',
    'grid_type1' / cs.Int32ul *
    'Type of 2nd grid used to derive a 2nd set of thickness data.',
    'grid_offset1' / cs.Int32ul *
    'File offset of the 2nd thickness data set in the file.',
    'prog_id' / cs.PaddedString(34, 'ascii') *
    'Internal progression ID (zero terminated string). All scans of the '
    'same progression share this ID.',
    '__empty' / cs.Padding(1790) *
    'Spare bytes for future use. Initialized to 0.',
    'localizer' / Localizer * 'NIR localizer image',
    'bscans' / bscan_format[cs.this.n_bscans],
)

# PR1 and EZ map to 14 and PR2 and IZ map to 15. Hence both names can be used
# to access the same data
SEG_MAPPING = {
    'ILM': 0,
    'BM': 1,
    'RNFL': 2,
    'GCL': 3,
    'IPL': 4,
    'INL': 5,
    'OPL': 6,
    'ONL': 7,
    'ELM': 8,
    'IOS': 9,
    'OPT': 10,
    'CHO': 11,
    'VIT': 12,
    'ANT': 13,
    'PR1': 14,
    'PR2': 15,
    'RPE': 16,
    'IPL+': 17,
    'IPL-': 18,
}

SLAB_MAPPING = {
    'NFLVP': ('ILM', 'RNFL'),
    'SVP': ('RNFL', 'IPL-'),
    'ICP': ('IPL-', 'IPL+'),
    'DCP': ('IPL+', 'OPL'),
    'SVC': ('ILM', 'IPL-'),
    'DVC': ('IPL-', 'OPL'),
    'AVAC': ('OPL', 'BM'), # Avascular Complex
    'RET': ('ILM', 'BM'), # Retina
}

TYPE_INTENSITY_MAPPING = {
    'oct': 'vol',
    'octa': 'angio',
}


class HeVolReader:

    def __init__(self, path, type: Literal['oct', 'octa'] = 'oct'):
        self.path = path
        self.intensity_transform = TYPE_INTENSITY_MAPPING.get(type, 'vol')
        with open(self.path, 'rb') as vol_file:
            self.parsed_file = vol_format.parse_stream(vol_file)

    @property
    def volume(self) -> EyeVolume:
        ## Check if scan pattern is supported by EyeVolume
        if self.parsed_file.scan_pattern == 2:
            msg = f'The EyeVolume object does not support scan pattern 2 (one Circular B-scan).'
            raise ValueError(msg)
        elif self.parsed_file.scan_pattern == 5:
            msg = f'The EyeVolume object does not support scan pattern 5 (Radial scan - star pattern).'
            raise ValueError(msg)

        data = np.stack([bscan.data for bscan in self.parsed_file.bscans],
                        axis=0)
        volume_meta = self.meta
        localizer = self.localizer
        volume = EyeVolume(
            data=data,
            meta=volume_meta,
            localizer=localizer,
            transformation=_compute_localizer_oct_transform(
                volume_meta, localizer.meta, data.shape),
        )

        layer_height_maps = self.layers
        for name, i in SEG_MAPPING.items():
            if i >= layer_height_maps.shape[0]:
                logger.warning(
                    'The volume contains less layers than expected. The naming might not be correct.'
                )
                break
            volume.add_layer_annotation(layer_height_maps[i], name=name)

        for name, (top, bottom) in SLAB_MAPPING.items():
            slab_meta = {
                'name': name,
                'top_layer': top,
                'bottom_layer': bottom
            }
            volume.add_slab_annotation(meta=slab_meta)

        return volume

    @property
    def layers(self):
        la = [b.layer_segmentations for b in self.parsed_file.bscans]
        n_layers = np.unique([len(l) for l in la])
        if len(n_layers) > 1:
            max_layers = np.max(n_layers)
            la = [
                np.pad(l, ((0, max_layers - len(l)), (0, 0)),
                       'constant',
                       constant_values=np.nan) for l in la
            ]
        layers = np.stack(la, axis=0)
        layers[layers >= 3.0e+38] = np.nan
        # Currently the shape is (n_bscans, n_layers, width). Swap the first two axes
        # to get (n_layers, n_bscans, width)
        return np.swapaxes(layers, 0, 1)

    @property
    def localizer_meta(self) -> EyeEnfaceMeta:
        return EyeEnfaceMeta(
            scale_x=self.parsed_file.scale_x_slo,
            scale_y=self.parsed_file.scale_y_slo,
            scale_unit='mm',
            modality='NIR',
            laterality=self.parsed_file.scan_position,
            field_size=self.parsed_file.field_size_slo,
            scan_focus=self.parsed_file.scan_focus,
            visit_date=self.parsed_file.visit_date,
            exam_time=self.parsed_file.exam_time,
        )

    @property
    def localizer(self) -> EyeEnface:
        localizer = self.parsed_file.localizer
        return EyeEnface(localizer, self.localizer_meta)

    @property
    def bscan_meta(self) -> list[EyeBscanMeta]:
        return [
            EyeBscanMeta(
                quality=b.quality,
                start_pos=(b.start_x, b.start_y),
                end_pos=(b.end_x, b.end_y),
                pos_unit='mm',
            ) for b in self.parsed_file.bscans
        ]

    @property
    def meta(self) -> EyeVolumeMeta:
        bscan_meta = self.bscan_meta
        return EyeVolumeMeta(
            scale_x=self.parsed_file.scale_x,
            scale_y=self.parsed_file.scale_y,
            scale_z=self.parsed_file.distance,
            scale_unit='mm',
            laterality=self.parsed_file.scan_position,
            visit_date=self.parsed_file.visit_date,
            exam_time=self.parsed_file.exam_time,
            bscan_meta=bscan_meta,
            intensity_transform=self.intensity_transform,
            par_algorithm='sum',
        )


class HeVolWriter:

    def __init__(self, volume: EyeVolume):
        self.volume = volume

    def write(self, path):
        with open(path, 'wb') as mf:
            mf.write(self.bytes)

    @property
    def bytes(self):
        return vol_format.build(
            dict(size_x=self.volume.size_x,
                 size_y=self.volume.size_y,
                 n_bscans=self.volume.size_z,
                 scale_x=self.volume.scale_x,
                 scale_y=self.volume.scale_y,
                 distance=self.volume.scale_z,
                 size_x_slo=self.volume.localizer.size_x,
                 size_y_slo=self.volume.localizer.size_y,
                 scale_x_slo=self.volume.localizer.scale_x,
                 scale_y_slo=self.volume.localizer.scale_y,
                 field_size_slo=self.volume.localizer.meta['field_size'],
                 scan_focus=self.volume.localizer.meta['scan_focus'],
                 scan_position=self.volume.localizer.meta['laterality'],
                 exam_time=datetime.datetime.combine(
                     self.volume.meta['exam_time'], datetime.time()),
                 scan_pattern=0,
                 id='unknown',
                 reference_id='unknown',
                 pid=self.volume.meta['patient']['pid'],
                 patient_id=self.volume.meta['patient']['patient_id'],
                 dob=datetime.datetime.combine(
                     self.volume.meta['patient']['dob'], datetime.time()),
                 vid=0,
                 visit_id='unknown',
                 visit_date=datetime.datetime.combine(
                     self.volume.meta['visit_date'], datetime.time()),
                 grid_type=0,
                 grid_offset=0,
                 grid_type1=0,
                 grid_offset1=0,
                 prog_id='unknown',
                 localizer=skimage.util.img_as_ubyte(
                     self.volume.localizer.data),
                 bscans=self._bscan_dicts))

    @property
    def _bscan_dicts(self):
        return [
            dict(bscan_hdr_size=256 + 17 * 4 * bscan.volume.size_x,
                 start_x=bscan.meta['start_pos'][0],
                 start_y=bscan.meta['start_pos'][1],
                 end_x=bscan.meta['end_pos'][0],
                 end_y=bscan.meta['end_pos'][1],
                 num_seg=len(bscan.layers),
                 quality=bscan.meta['quality'],
                 shift=0,
                 iv_transformation=[0, 0, 0, 0, 0, 0],
                 layer_segmentations=self._segmentations_from_bscan(bscan),
                 data=to_vol_intensity(bscan.data.astype(np.float32)))
            for bscan in self.volume
        ]

    def _segmentations_from_bscan(self, bscan: EyeBscan) -> np.ndarray:
        """"""
        segs = np.zeros((17, bscan.volume.size_x), dtype=np.float32)
        for layer in bscan.layers:
            if layer.name in SEG_MAPPING:
                segs[SEG_MAPPING[layer.name]] = layer.data
            else:
                warnings.warn(
                    f'Unknown layer name: {layer.name}. Skipping. \n VOL format only supports layers with the following names: {list(SEG_MAPPING.keys())}'
                )
        return segs



from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from eyepy.quant.metrics import compute_area
from eyepy.quant.spatial import _normalize_origin_mode
from eyepy.quant.spatial import AnatomicalOrigin
from eyepy.quant.spatial import DirectionalExtent
from eyepy.quant.spatial import OriginMode
from eyepy.quant.spatial import OriginModeType
from eyepy.quant.spatial import PolarReference

if TYPE_CHECKING:
    from matplotlib.axes import Axes

    from eyepy.core.eyeenface import EyeEnface


@dataclass
class RegionQuantification:
    """Base class for region quantification results.

    Attributes:
        area: Area of the region in square micrometers
        extent: Directional extent from anatomical origin
        origin: Anatomical origin used for directional measurements (optional, for plotting)
        mask: Binary mask of the region (optional, for plotting)
        scale_x: Micrometers per pixel in x-direction (optional, for plotting)
        scale_y: Micrometers per pixel in y-direction (optional, for plotting)
        unit: Unit for displaying distances, has to match the provided scale (Default: 'px')
    """

    area: float
    extent: DirectionalExtent
    origin: Optional[AnatomicalOrigin] = None
    scale_x: float = 1.0
    scale_y: float = 1.0
    unit: str = 'px'

    def to_dict(self) -> dict:
        """Convert to dictionary for easy export.

        Returns:
            Dictionary with all quantification metrics
        """
        result = {
            'area': self.area,
        }
        result.update(self.extent.to_dict())
        return result

    def _format_distance(self, distance: float) -> str:
        """Format distance value according to the unit property.

        Args:
            distance: Distance value in micrometers

        Returns:
            Formatted string with value and unit
        """
        return f'{distance:.1f} {self.unit}'

    def plot(
        self,
        ax: Optional['Axes'] = None,
        show_origin: bool = True,
        show_extent_lines: bool = True,
        directions: Optional[list[str]] = None,
        origin_color: str = 'blue',
        origin_marker: str = 'x',
        origin_size: float = 100,
        line_color: str = 'yellow',
        line_width: float = 2,
        font_size: int = 10,
    ) -> 'Axes':
        """Plot the region with directional extent measurements.

        Visualizes the quantification by showing:
        - The anatomical origin point
        - Directional extent lines with distance annotations

        Args:
            ax: Matplotlib axes to plot on (creates new if None)
            show_origin: Whether to mark the origin point
            show_extent_lines: Whether to show directional extent lines
            directions: List of directions to plot. If None, plots all 8 directions.
                       Valid values: 'temporal', 'nasal', 'superior', 'inferior',
                       'superior_temporal', 'inferior_temporal', 'superior_nasal',
                       'inferior_nasal'
            origin_color: Color for the origin marker
            origin_marker: Marker style for origin point
            origin_size: Size of origin marker
            line_color: Color for extent lines
            line_width: Width of extent lines
            font_size: Font size for distance annotations

        Returns:
            Matplotlib axes with the plot

        Raises:
            ValueError: If origin is not available when needed
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 10))
        ax.set_aspect('equal')

        # Check if origin is available for remaining visualizations
        if (show_origin or show_extent_lines) and self.origin is None:
            raise ValueError(
                'Origin must be provided to plot origin point or extent lines. '
                'Use from_mask() or from_mask_with_origin() to create quantification '
                'with origin information.'
            )

        # Show origin point
        if show_origin and self.origin is not None:
            ax.scatter(
                self.origin.x,
                self.origin.y,
                c=origin_color,
                marker=origin_marker,
                s=origin_size,
                zorder=10,
                label='Origin',
            )

        # Show directional extent lines
        if show_extent_lines and self.origin is not None:
            # Define all available directions and their angles
            all_directions = {
                'temporal': (0, 'T'),
                'nasal': (np.pi, 'N'),
                'superior': (-np.pi / 2, 'S'),
                'inferior': (np.pi / 2, 'I'),
                'superior_temporal': (-np.pi / 4, 'ST'),
                'inferior_temporal': (np.pi / 4, 'IT'),
                'superior_nasal': (-3 * np.pi / 4, 'SN'),
                'inferior_nasal': (3 * np.pi / 4, 'IN'),
            }

            # Filter to requested directions (or all if None)
            if directions is None:
                directions_to_plot = all_directions
            else:
                # Validate requested directions
                invalid = set(directions) - set(all_directions.keys())
                if invalid:
                    raise ValueError(
                        f'Invalid direction(s): {invalid}. '
                        f'Valid directions are: {list(all_directions.keys())}'
                    )
                directions_to_plot = {k: v for k, v in all_directions.items() if k in directions}

            for direction, (angle, label) in directions_to_plot.items():
                extent_metrics = getattr(self.extent, direction, None)
                if extent_metrics is None:
                    continue

                # Extract midpoint distance from ExtentMetrics
                distance = extent_metrics.midpoint
                if distance == 0:
                    continue

                # Convert angle and distance to Cartesian endpoint
                # Note: in image coordinates, y increases downward
                # angle=0 is temporal (positive x_cart)
                # We need to convert from anatomical coordinates back to image coordinates
                x_cart = distance * np.cos(angle)
                y_cart = distance * np.sin(angle)

                # Convert back to image coordinates from anatomical coordinates
                # For OD: x_cart positive = temporal = left = negative dx_image
                # For OS: x_cart positive = temporal = right = positive dx_image
                if self.origin.laterality == 'OD':
                    dx_image = -x_cart / self.scale_x
                else:
                    dx_image = x_cart / self.scale_x

                dy_image = y_cart / self.scale_y

                end_x = self.origin.x + dx_image
                end_y = self.origin.y + dy_image

                # Draw line from origin to endpoint
                ax.plot(
                    [self.origin.x, end_x],
                    [self.origin.y, end_y],
                    color=line_color,
                    linewidth=line_width,
                    zorder=5,
                )

                # Add text annotation at midpoint
                mid_x = (self.origin.x + end_x) / 2
                mid_y = (self.origin.y + end_y) / 2

                # Format distance based on unit
                dist_text = self._format_distance(distance)

                ax.text(
                    mid_x,
                    mid_y,
                    f'{label}: {dist_text}',
                    color='white',
                    fontsize=font_size,
                    ha='center',
                    va='center',
                    bbox=dict(
                        boxstyle='round,pad=0.3',
                        facecolor='black',
                        alpha=0.7,
                        edgecolor='none',
                    ),
                    zorder=15,
                )

        return ax

    @classmethod
    def from_mask_with_origin(
        cls,
        mask: npt.NDArray[np.bool_],
        origin: AnatomicalOrigin,
        scale_x: float = 1.0,
        scale_y: float = 1.0,
        unit: str = 'px',
    ):
        """Quantify region from a binary mask with custom origin.

        Computes directional extent (midpoint) and arc statistics in all 8 directions (temporal, nasal,
        superior, inferior, superior_temporal, inferior_temporal, superior_nasal,
        inferior_nasal).

        Args:
            mask: Binary mask of the region
            origin: AnatomicalOrigin defining the reference point
            scale_x: Size of a pixel in x-direction (default: 1.0)
            scale_y: Size of a pixel in y-direction (default: 1.0)
            unit: Unit for displaying distances (default: 'px')

        Returns:
            RegionQuantification subclass instance with computed metrics
        """
        # Compute area
        area = compute_area(mask, scale_x, scale_y)

        # Compute directional extent
        polar_ref = PolarReference(origin)
        extent = polar_ref.compute_directional_extent(
            mask,
            scale_x=scale_x,
            scale_y=scale_y,
        )

        return cls(
            area=area,
            extent=extent,
            origin=origin,
            scale_x=scale_x,
            scale_y=scale_y,
            unit=unit,
        )

    @classmethod
    def from_mask(
        cls,
        mask: npt.NDArray[np.bool_],
        optic_disc_center: Optional[tuple[float, float]] = None,
        fovea_center: Optional[tuple[float, float]] = None,
        laterality: Optional[str] = None,
        scale_x: float = 1.0,
        scale_y: float = 1.0,
        origin_mode: OriginModeType = OriginMode.HYBRID,
        unit: str = 'px',
    ):
        """Quantify region from a binary mask.

        Computes directional extent (midpoint) and arc statistics in all 8 directions (temporal, nasal,
        superior, inferior, superior_temporal, inferior_temporal, superior_nasal,
        inferior_nasal) using both midpoint and statistics methods.

        Args:
            mask: Binary mask of the region
            optic_disc_center: (y, x) coordinates of optic disc center
            fovea_center: (y, x) coordinates of fovea center
            laterality: Eye laterality ('OD' or 'OS')
            scale_x: Micrometers per pixel in x-direction
            scale_y: Micrometers per pixel in y-direction
            origin_mode: Mode determining the reference origin.
                        Can be 'optic_disc', 'fovea', 'hybrid' or 'custom'
                        (default: 'hybrid')
            unit: Unit for displaying distances (default: 'px')

        Returns:
            RegionQuantification subclass instance with computed metrics
        """
        # Normalize origin_mode to ENUM
        origin_mode = _normalize_origin_mode(origin_mode)

        # Create anatomical origin based on mode
        if origin_mode == OriginMode.OPTIC_DISC:
            if optic_disc_center is None:
                raise ValueError(
                    'optic_disc_center must be provided when origin_mode is OPTIC_DISC.'
                )
            if laterality is None:
                raise ValueError(
                    'laterality must be provided when origin_mode is OPTIC_DISC.'
                )
            origin = AnatomicalOrigin.from_optic_disc(
                optic_disc_center=optic_disc_center,
                laterality=laterality,
            )
        elif origin_mode == OriginMode.FOVEA:
            if fovea_center is None:
                raise ValueError(
                    'fovea_center must be provided when origin_mode is FOVEA.'
                )
            if laterality is None:
                raise ValueError(
                    'laterality must be provided when origin_mode is FOVEA.'
                )
            origin = AnatomicalOrigin.from_fovea(
                fovea_center=fovea_center,
                laterality=laterality,
            )
        elif origin_mode == OriginMode.HYBRID:
            if optic_disc_center is None:
                raise ValueError(
                    'optic_disc_center must be provided when origin_mode is HYBRID.'
                )
            if fovea_center is None:
                raise ValueError(
                    'fovea_center must be provided when origin_mode is HYBRID.'
                )
            if laterality is None:
                raise ValueError(
                    'laterality must be provided when origin_mode is HYBRID.'
                )

            origin = AnatomicalOrigin.from_hybrid(
                optic_disc_center=optic_disc_center,
                fovea_center=fovea_center,
                laterality=laterality,
            )
        elif origin_mode == OriginMode.CUSTOM:
            raise ValueError(
                f'Use from_mask_with_origin() for OriginMode.CUSTOM'
            )

        return cls.from_mask_with_origin(
            mask=mask,
            origin=origin,
            scale_x=scale_x,
            scale_y=scale_y,
            unit=unit,
        )

    @classmethod
    def from_EyeEnface(
        cls,
        enface: 'EyeEnface',
        area_map_name: str,
        origin_mode: OriginModeType = OriginMode.HYBRID,
        custom_origin: Optional[AnatomicalOrigin] = None,
    ):
        """Create RegionQuantification from an EyeEnface object.

        A RegionQuantification quantifies an annotation mask of one connected component.
        Extents from a specified origin are computed in 8 directions.

        Args:
            enface: EyeEnface object containing the image and annotations
            area_map_name: Name of the area map containing the region to quantify
            origin_mode: Mode determining the reference origin.
                        Can be 'optic_disc', 'fovea', 'hybrid' or 'custom'
                        (default: 'hybrid')
            custom_origin: AnatomicalOrigin to use when origin_mode is 'custom'
                          (default: None)

        Returns:
            RegionQuantification subclass instance with computed metrics

        Raises:
            ValueError: If required annotations or area map are missing
        """
        # Normalize origin_mode to ENUM
        origin_mode = _normalize_origin_mode(origin_mode)

        # Validate based on origin_mode
        if origin_mode == OriginMode.CUSTOM:
            if custom_origin is None:
                raise ValueError(
                    'custom_origin must be provided when origin_mode is CUSTOM.'
                )
        elif origin_mode == OriginMode.OPTIC_DISC:
            if enface.optic_disc is None:
                raise ValueError(
                    'EyeEnface must have optic_disc annotation when origin_mode is OPTIC_DISC. '
                    'Set enface.optic_disc before calling from_EyeEnface().'
                )
        elif origin_mode == OriginMode.FOVEA:
            if enface.fovea is None:
                raise ValueError(
                    'EyeEnface must have fovea annotation when origin_mode is FOVEA. '
                    'Set enface.fovea before calling from_EyeEnface().'
                )
        elif origin_mode == OriginMode.HYBRID:
            if enface.optic_disc is None:
                raise ValueError(
                    'EyeEnface must have optic_disc annotation when origin_mode is HYBRID. '
                    'Set enface.optic_disc before calling from_EyeEnface().'
                )
            if enface.fovea is None:
                raise ValueError(
                    'EyeEnface must have fovea annotation when origin_mode is HYBRID. '
                    'Set enface.fovea before calling from_EyeEnface().'
                )

        # Get the area map
        if area_map_name not in enface.area_maps:
            raise ValueError(
                f'Area map "{area_map_name}" not found in EyeEnface. '
                f'Available area maps: {list(enface.area_maps.keys())}.'
            )

        area_annotation = enface.area_maps[area_map_name]
        mask = area_annotation.data

        # Get scales
        scale_x = enface.scale_x
        scale_y = enface.scale_y

        # Get unit from metadata if available, otherwise default to 'px'
        unit = enface.meta.get('scale_unit', 'px')

        # Handle different origin modes
        if origin_mode == OriginMode.CUSTOM:
            # custom_origin is guaranteed to be not None due to validation above
            assert custom_origin is not None
            return cls.from_mask_with_origin(
                mask=mask,
                origin=custom_origin,
                scale_x=scale_x,
                scale_y=scale_y,
                unit=unit,
            )
        else:
            # Check for laterality (needed for non-CUSTOM modes)
            laterality = enface.meta.get('laterality', None)
            if laterality is None:
                raise ValueError(
                    'EyeEnface must have laterality information in metadata '
                    f'when origin_mode is {origin_mode.value}. '
                    'Set enface.meta["laterality"] to "OD" or "OS".'
                )

            # Get anatomical landmarks (guaranteed to be not None due to validation above)
            optic_disc_center = enface.optic_disc.center if enface.optic_disc else (0.0, 0.0)
            fovea_center = enface.fovea.center if enface.fovea else (0.0, 0.0)

            # Create quantification using existing from_mask method
            return cls.from_mask(
                mask=mask,
                optic_disc_center=optic_disc_center,
                fovea_center=fovea_center,
                laterality=laterality,
                scale_x=scale_x,
                scale_y=scale_y,
                origin_mode=origin_mode,
                unit=unit,
            )



import logging
from pathlib import Path
import urllib.request
import zipfile

import eyepy as ep

logger = logging.getLogger(__name__)

SAMPLE_DATA = {
    'drusen_patient': (
        'https://github.com/MedVisBonn/eyepydata/releases/download/v1.0.0/drusen_patient.zip',
        ep.import_heyex_xml,
    ),
    'healthy_OS': (
        'https://github.com/MedVisBonn/eyepydata/releases/download/v1.0.0/healthy_OS.vol',
        ep.import_heyex_vol,
    ),
    'healthy_OD': (
        'https://github.com/MedVisBonn/eyepydata/releases/download/v1.0.0/healthy_OD.vol',
        ep.import_heyex_vol,
    ),
    'healthy_OS_Angio': (
        'https://github.com/MedVisBonn/eyepydata/releases/download/v1.0.0/healthy_OS_Angio.vol',
        ep.import_heyex_angio_vol,
    ),
    'healthy_OD_Angio': (
        'https://github.com/MedVisBonn/eyepydata/releases/download/v1.0.0/healthy_OD_Angio.vol',
        ep.import_heyex_angio_vol,
    ),
}

EYEPY_DATA_DIR = Path('~/.eyepy/data').expanduser()
if not EYEPY_DATA_DIR.is_dir():
    EYEPY_DATA_DIR.mkdir(parents=True)


def load(name: str) -> ep.EyeVolume:
    """Load sample data.

    Args:
        name: Name of the sample data to load.
            Available options are:
            - 'drusen_patient': Sample data with many drusen.
            - 'healthy_OS': Healthy left eye volume.
            - 'healthy_OD': Healthy right eye volume.
            - 'healthy_OS_Angio': Healthy left eye OCT Angiography.
            - 'healthy_OD_Angio': Healthy right eye OCT Angiography.

    Returns:
        ep.EyeVolume: The loaded eye volume data.
    """
    url, import_func = SAMPLE_DATA[name]
    file_ext = Path(url).suffix
    data_path = EYEPY_DATA_DIR / name

    if file_ext == '.zip':
        # ZIP file: extract if not already extracted
        if not data_path.is_dir():
            download_path = EYEPY_DATA_DIR / (name + '.zip')
            urllib.request.urlretrieve(url, download_path)
            with zipfile.ZipFile(download_path, 'r') as zip_ref:
                zip_ref.extractall(EYEPY_DATA_DIR)
            download_path.unlink()
        # Assume XML inside extracted folder for zip samples
        return import_func(data_path / 'metaclean.xml')
    else:
        # Direct file (e.g., .vol): download if not present
        if not data_path.exists():
            urllib.request.urlretrieve(url, data_path)
        return import_func(data_path)



"""Spatial reference systems for anatomical quantification.

This module provides classes for defining spatial reference frames based
on anatomical landmarks (optic disc, fovea) and computing distances and
directions relative to these references.
"""

from dataclasses import dataclass
from enum import Enum
from typing import cast, Literal, Optional, Union

import numpy as np
import numpy.typing as npt


class OriginMode(Enum):
    """Mode for determining the anatomical origin.

    OPTIC_DISC: Use optic disc center as origin
    FOVEA: Use fovea center as origin
    HYBRID: Use optic disc x-coordinate and fovea y-coordinate
    CUSTOM: Use a custom user-specified origin point
    """

    OPTIC_DISC = 'optic_disc'
    FOVEA = 'fovea'
    HYBRID = 'hybrid'
    CUSTOM = 'custom'


# Type alias for flexible API - accept both ENUM and strings
OriginModeType = Union[OriginMode, Literal['optic_disc', 'fovea', 'hybrid', 'custom']]


def _normalize_origin_mode(mode: OriginModeType) -> OriginMode:
    """Convert string to OriginMode ENUM if needed.

    Args:
        mode: Either an OriginMode enum or a string literal

    Returns:
        OriginMode enum

    Raises:
        ValueError: If string mode is not a valid option
    """
    if isinstance(mode, str):
        try:
            return OriginMode(mode)
        except ValueError:
            valid_values = [m.value for m in OriginMode]
            raise ValueError(
                f"Invalid origin_mode: '{mode}'. "
                f'Must be one of {valid_values} or use OriginMode enum'
            )
    return mode


@dataclass
class AnatomicalOrigin:
    """Reference origin based on anatomical landmarks.

    Defines a coordinate system origin that can be based on the optic disc,
    fovea, a hybrid approach, or a custom position.

    **Coordinate System Convention:**
    This class uses (row, col) image coordinates for input and output:
    - row: vertical axis, increases downward (corresponds to y)
    - col: horizontal axis, increases rightward (corresponds to x)

    The internal storage uses (y, x) where y=row and x=col.
    All methods that accept or return coordinates use (row, col) format unless
    explicitly documented otherwise.

    Attributes:
        y: Vertical (y) coordinate of the origin
        x: Horizontal (x) coordinate of the origin
        laterality: Eye laterality ('OD' or 'OS')
        mode: Origin mode used to determine the position
    """

    y: float
    x: float
    laterality: str
    mode: OriginMode

    @classmethod
    def from_optic_disc(
        cls,
        optic_disc_center: tuple[float, float],
        laterality: str,
    ) -> 'AnatomicalOrigin':
        """Create origin at optic disc center.

        Args:
            optic_disc_center: (y, x) coordinates of optic disc center
            laterality: Eye laterality ('OD' or 'OS')

        Returns:
            AnatomicalOrigin at optic disc center
        """
        if laterality not in ['OD', 'OS']:
            raise ValueError(f'Laterality must be OD or OS, got {laterality}')

        return cls(
            y=optic_disc_center[0],
            x=optic_disc_center[1],
            laterality=laterality,
            mode=OriginMode.OPTIC_DISC,
        )

    @classmethod
    def from_fovea(
        cls,
        fovea_center: tuple[float, float],
        laterality: str,
    ) -> 'AnatomicalOrigin':
        """Create origin at fovea center.

        Args:
            fovea_center: (y, x) coordinates of fovea center
            laterality: Eye laterality ('OD' or 'OS')

        Returns:
            AnatomicalOrigin at fovea center
        """
        if laterality not in ['OD', 'OS']:
            raise ValueError(f'Laterality must be OD or OS, got {laterality}')

        return cls(
            y=fovea_center[0],
            x=fovea_center[1],
            laterality=laterality,
            mode=OriginMode.FOVEA,
        )

    @classmethod
    def from_hybrid(
        cls,
        optic_disc_center: tuple[float, float],
        fovea_center: tuple[float, float],
        laterality: str,
    ) -> 'AnatomicalOrigin':
        """Create hybrid origin from optic disc and fovea positions.

        Uses the horizontal (x) position from the optic disc center and
        the vertical (y) position from the fovea center.

        Args:
            optic_disc_center: (y, x) coordinates of optic disc center
            fovea_center: (y, x) coordinates of fovea center
            laterality: Eye laterality ('OD' or 'OS')

        Returns:
            AnatomicalOrigin with y from fovea, x from optic disc
        """
        if laterality not in ['OD', 'OS']:
            raise ValueError(f'Laterality must be OD or OS, got {laterality}')

        # Origin: horizontal position from OD, vertical position from fovea
        origin_y = fovea_center[0]
        origin_x = optic_disc_center[1]

        return cls(
            y=origin_y,
            x=origin_x,
            laterality=laterality,
            mode=OriginMode.HYBRID,
        )

    @classmethod
    def from_custom(
        cls,
        origin: tuple[float, float],
        laterality: str,
    ) -> 'AnatomicalOrigin':
        """Create origin at custom position.

        Args:
            origin: (y, x) coordinates of custom origin
            laterality: Eye laterality ('OD' or 'OS')

        Returns:
            AnatomicalOrigin at custom position
        """
        if laterality not in ['OD', 'OS']:
            raise ValueError(f'Laterality must be OD or OS, got {laterality}')

        return cls(
            y=origin[0],
            x=origin[1],
            laterality=laterality,
            mode=OriginMode.CUSTOM,
        )

    def to_cartesian(self, y: float, x: float) -> tuple[float, float]:
        """Convert image coordinates (row, col) to Cartesian coordinates relative
        to origin.

        Image coordinates use (row, col) convention where:
        - row increases downward
        - col increases to the right

        Cartesian output coordinates are anatomically oriented:
        - x: horizontal (positive = temporal, negative = nasal)
        - y: vertical (positive = inferior, negative = superior)

        Args:
            y: Image row coordinate (increases downward)
            x: Image column coordinate (increases to the right)

        Returns:
            (x_cart, y_cart) Cartesian coordinates relative to origin where:
            - x_cart: horizontal distance (positive = temporal, negative = nasal)
            - y_cart: vertical distance (positive = inferior/downward, negative = superior/upward)
        """
        # Compute displacement in image coordinates
        dx_image = x - self.x  # Column displacement
        dy_image = y - self.y  # Row displacement (positive = downward)

        # In Cartesian coordinates:
        # - y_cart = dy_image (positive downward = inferior, negative upward = superior)
        # - x_cart needs laterality adjustment for temporal/nasal

        y_cart = dy_image  # Positive = inferior/downward, negative = superior/upward

        # Adjust for laterality: temporal direction
        # OD (right eye): nasal is to the right (positive dx), temporal is to the left (negative dx)
        #                 So temporal (positive x_cart) = negative dx_image
        # OS (left eye): nasal is to the left (negative dx), temporal is to the right (positive dx)
        #                So temporal (positive x_cart) = positive dx_image
        if self.laterality == 'OD':
            x_cart = -dx_image
        else:  # OS
            x_cart = dx_image

        return (x_cart, y_cart)

    def to_polar(self, y: float, x: float) -> tuple[float, float]:
        """Convert image coordinates (row, col) to polar coordinates.

        Args:
            y: Image row coordinate (increases downward)
            x: Image column coordinate (increases to the right)

        Returns:
            (distance, angle) where:
            - distance: Euclidean distance from origin
            - angle: Angle in radians (0 = temporal, π/2 = inferior,
                     π = nasal, 3π/2 = superior)
        """
        x_cart, y_cart = self.to_cartesian(y, x)
        distance = np.sqrt(x_cart**2 + y_cart**2)
        angle = np.arctan2(y_cart, x_cart)  # Range: [-π, π]

        # Convert to [0, 2π] with 0 = temporal
        if angle < 0:
            angle += 2 * np.pi

        return (distance, angle)


@dataclass
class ExtentMetrics:
    """Metrics for extent in a single direction.

    Contains statistical measures of distances from the origin to boundary
    points within a single angular sector.

    Attributes:
        midpoint: Distance to boundary at exact midpoint angle of this direction
        mean: Mean distance to boundary points in this direction
        max: Maximum distance to boundary in this direction
        median: Median distance to boundary points in this direction
        std: Standard deviation of distances in this direction
        touches_border: Whether the region extends to the image border in this
                       specific direction, indicating measurements may be truncated
    """

    midpoint: float
    mean: float
    max: float
    median: float
    std: float
    touches_border: bool = False

    def to_dict(self) -> dict:
        """Convert to dictionary.

        Returns:
            Dictionary mapping metric names to values
        """
        return {
            'midpoint': self.midpoint,
            'mean': self.mean,
            'max': self.max,
            'median': self.median,
            'std': self.std,
            'touches_border': self.touches_border,
        }


@dataclass
class DirectionalExtent:
    """Extent of a region in specific directions from origin.

    Contains ExtentMetrics for each of the 8 anatomical directions (4 cardinal
    + 4 ordinal) from an anatomical origin. Each direction includes mean, max,
    median, and standard deviation of boundary distances, along with a flag
    indicating if that specific direction touches the image border.

    Attributes:
        temporal: Extent metrics in temporal direction
        nasal: Extent metrics in nasal direction
        superior: Extent metrics in superior direction
        inferior: Extent metrics in inferior direction
        superior_temporal: Extent metrics in superior-temporal direction
        inferior_temporal: Extent metrics in inferior-temporal direction
        superior_nasal: Extent metrics in superior-nasal direction
        inferior_nasal: Extent metrics in inferior-nasal direction
    """

    temporal: ExtentMetrics
    nasal: ExtentMetrics
    superior: ExtentMetrics
    inferior: ExtentMetrics
    superior_temporal: ExtentMetrics
    inferior_temporal: ExtentMetrics
    superior_nasal: ExtentMetrics
    inferior_nasal: ExtentMetrics

    def to_dict(self) -> dict:
        """Convert to dictionary.

        Returns:
            Dictionary with nested structure: direction -> metric -> value
        """
        return {
            'temporal': self.temporal.to_dict(),
            'nasal': self.nasal.to_dict(),
            'superior': self.superior.to_dict(),
            'inferior': self.inferior.to_dict(),
            'superior_temporal': self.superior_temporal.to_dict(),
            'inferior_temporal': self.inferior_temporal.to_dict(),
            'superior_nasal': self.superior_nasal.to_dict(),
            'inferior_nasal': self.inferior_nasal.to_dict(),
        }


def _extract_boundary_with_border(mask: npt.NDArray[np.bool_]) -> tuple[npt.NDArray, npt.NDArray, bool]:
    """Extract boundary pixels including those at image borders.

    This function properly handles masks that extend to the image border,
    ensuring the outer boundary is captured even when the region touches edges.

    Args:
        mask: Binary mask of the region

    Returns:
        y_coords: Y coordinates of boundary pixels
        x_coords: X coordinates of boundary pixels
        touches_border: Whether the region touches the image border
    """
    from scipy import ndimage
    from skimage import morphology

    # Fill holes to get outer boundary only (ignore inner boundaries)
    filled_mask = ndimage.binary_fill_holes(mask)
    assert filled_mask is not None  # Type narrowing

    # Check if region touches the border
    h, w = filled_mask.shape
    touches_border = (
        np.any(filled_mask[0, :]) or  # Top edge
        np.any(filled_mask[-1, :]) or  # Bottom edge
        np.any(filled_mask[:, 0]) or  # Left edge
        np.any(filled_mask[:, -1])     # Right edge
    )

    if touches_border:
        # Pad the mask to avoid losing border pixels during erosion
        padded_mask = np.pad(filled_mask, pad_width=1, mode='constant', constant_values=False)

        # Extract boundary from padded mask
        eroded = morphology.binary_erosion(padded_mask)
        boundary_padded = padded_mask & ~eroded

        # Remove padding (shift coordinates back)
        boundary = boundary_padded[1:-1, 1:-1]
    else:
        # Normal case: erosion works correctly
        eroded = morphology.binary_erosion(filled_mask)
        boundary = filled_mask & ~eroded

    # Get coordinates of boundary pixels
    y_coords, x_coords = np.where(boundary)

    return y_coords, x_coords, bool(touches_border)


class PolarReference:
    """Polar coordinate reference system for spatial analysis.

    Divides the image into 8 angular sectors (4 cardinal + 4 ordinal directions)
    relative to an anatomical origin for computing directional statistics.

    Attributes:
        origin: Anatomical origin point
    """

    def __init__(
        self,
        origin: AnatomicalOrigin,
    ):
        """Initialize polar reference system.

        Args:
            origin: Anatomical origin defining the coordinate system
        """
        self.origin = origin

    def compute_directional_extent(
        self,
        mask: npt.NDArray[np.bool_],
        scale_x: float = 1.0,
        scale_y: float = 1.0,
    ) -> DirectionalExtent:
        """Compute extent in all 8 directions with complete metrics.

        For each of the 8 anatomical directions (4 cardinal + 4 ordinal),
        computes:
        - Midpoint distance: distance to boundary at exact direction angle
        - Mean, max, median, std: statistics of all boundary points in that sector
        - Border flag: whether that specific direction touches the image border

        Args:
            mask: Binary mask of the region
            scale_x: Micrometers per pixel in x-direction
            scale_y: Micrometers per pixel in y-direction

        Returns:
            DirectionalExtent with ExtentMetrics for all 8 directions
        """
        # Define the 8 directions with their midpoint angles and sector ranges
        # Each sector spans π/4 radians (45 degrees) centered on its midpoint
        directions_config = {
            'temporal': {
                'midpoint_angle': 0.0,
                'sector_range': (-np.pi / 8, np.pi / 8),
            },
            'inferior_temporal': {
                'midpoint_angle': np.pi / 4,
                'sector_range': (np.pi / 8, 3 * np.pi / 8),
            },
            'inferior': {
                'midpoint_angle': np.pi / 2,
                'sector_range': (3 * np.pi / 8, 5 * np.pi / 8),
            },
            'inferior_nasal': {
                'midpoint_angle': 3 * np.pi / 4,
                'sector_range': (5 * np.pi / 8, 7 * np.pi / 8),
            },
            'nasal': {
                'midpoint_angle': np.pi,
                'sector_range': ((7 * np.pi / 8, np.pi), (-np.pi, -7 * np.pi / 8)),
            },
            'superior_nasal': {
                'midpoint_angle': -3 * np.pi / 4,
                'sector_range': (-7 * np.pi / 8, -5 * np.pi / 8),
            },
            'superior': {
                'midpoint_angle': -np.pi / 2,
                'sector_range': (-5 * np.pi / 8, -3 * np.pi / 8),
            },
            'superior_temporal': {
                'midpoint_angle': -np.pi / 4,
                'sector_range': (-3 * np.pi / 8, -np.pi / 8),
            },
        }

        if not np.any(mask):
            # Empty mask - return zeros for all directions
            zero_metrics = ExtentMetrics(
                midpoint=0.0, mean=0.0, max=0.0, median=0.0, std=0.0, touches_border=False
            )
            return DirectionalExtent(
                temporal=zero_metrics,
                nasal=zero_metrics,
                superior=zero_metrics,
                inferior=zero_metrics,
                superior_temporal=zero_metrics,
                inferior_temporal=zero_metrics,
                superior_nasal=zero_metrics,
                inferior_nasal=zero_metrics,
            )

        # Extract boundary pixels, properly handling border cases
        y_coords, x_coords, _ = _extract_boundary_with_border(mask)

        if len(y_coords) == 0:
            # Single pixel mask - use the pixel itself
            y_coords, x_coords = np.where(mask)

        # Compute metrics for all 8 directions
        mask_shape = (mask.shape[0], mask.shape[1])
        metrics = {}
        for direction_name, config in directions_config.items():
            metrics[direction_name] = self._compute_single_direction_metrics(
                y_coords=y_coords,
                x_coords=x_coords,
                scale_x=scale_x,
                scale_y=scale_y,
                midpoint_angle=config['midpoint_angle'],
                sector_range=config['sector_range'],
                mask_shape=mask_shape,
            )

        return DirectionalExtent(
            temporal=metrics['temporal'],
            nasal=metrics['nasal'],
            superior=metrics['superior'],
            inferior=metrics['inferior'],
            superior_temporal=metrics['superior_temporal'],
            inferior_temporal=metrics['inferior_temporal'],
            superior_nasal=metrics['superior_nasal'],
            inferior_nasal=metrics['inferior_nasal'],
        )

    def _compute_single_direction_metrics(
        self,
        y_coords: npt.NDArray,
        x_coords: npt.NDArray,
        scale_x: float,
        scale_y: float,
        midpoint_angle: float,
        sector_range: Union[tuple[float, float], tuple[tuple[float, float], tuple[float, float]]],
        mask_shape: tuple[int, int],
    ) -> ExtentMetrics:
        """Compute all metrics for a single direction.

        Args:
            y_coords: Y coordinates of boundary pixels
            x_coords: X coordinates of boundary pixels
            scale_x: Micrometers per pixel in x-direction
            scale_y: Micrometers per pixel in y-direction
            midpoint_angle: Exact angle (in radians) for midpoint measurement
            sector_range: Angular range for this sector (angle_min, angle_max)
                         or ((angle_min1, angle_max1), (angle_min2, angle_max2)) for wraparound
            mask_shape: Shape of the mask (height, width) for border detection

        Returns:
            ExtentMetrics with all computed metrics for this direction
        """
        # Collect distances and boundary points in this sector
        distances = []
        sector_boundary_points = []

        for y, x in zip(y_coords, x_coords):
            x_cart, y_cart = self.origin.to_cartesian(y, x)

            # Skip origin point
            if x_cart == 0 and y_cart == 0:
                continue

            # Compute angle
            angle = np.arctan2(y_cart, x_cart)  # Range: [-π, π]

            # Check if point is in this sector
            in_sector = False
            if isinstance(sector_range[0], tuple):
                # Wraparound case (e.g., nasal direction)
                # sector_range is ((angle_min1, angle_max1), (angle_min2, angle_max2))
                wraparound_ranges = cast(tuple[tuple[float, float], tuple[float, float]], sector_range)
                in_sector = (angle >= wraparound_ranges[0][0]) or (angle <= wraparound_ranges[1][1])
            else:
                # Normal case: sector_range is (angle_min, angle_max)
                simple_range = cast(tuple[float, float], sector_range)
                in_sector = simple_range[0] <= angle < simple_range[1]

            if in_sector:
                # Compute scaled distance
                x_cart_scaled = x_cart * scale_x
                y_cart_scaled = y_cart * scale_y
                distance = np.sqrt(x_cart_scaled**2 + y_cart_scaled**2)
                distances.append(distance)
                sector_boundary_points.append((y, x))

        # Compute statistics
        if len(distances) > 0:
            mean_dist = float(np.mean(distances))
            max_dist = float(np.max(distances))
            median_dist = float(np.median(distances))
            std_dist = float(np.std(distances))
        else:
            mean_dist = max_dist = median_dist = std_dist = 0.0

        # Compute midpoint distance (distance at exact midpoint angle)
        # Find boundary point in this sector closest to the midpoint angle
        midpoint_dist = 0.0
        # Use sector_boundary_points if available (normal case)
        points_to_search = sector_boundary_points
        if len(points_to_search) > 0:
            min_angle_diff = float('inf')
            for y, x in points_to_search:
                x_cart, y_cart = self.origin.to_cartesian(y, x)
                if x_cart == 0 and y_cart == 0:
                    continue

                angle = np.arctan2(y_cart, x_cart)
                angle_diff = abs(angle - midpoint_angle)

                # Handle wraparound at ±π
                if angle_diff > np.pi:
                    angle_diff = 2 * np.pi - angle_diff

                if angle_diff < min_angle_diff:
                    min_angle_diff = angle_diff
                    x_cart_scaled = x_cart * scale_x
                    y_cart_scaled = y_cart * scale_y
                    midpoint_dist = float(np.sqrt(x_cart_scaled**2 + y_cart_scaled**2))

        # Check if this direction touches the border
        touches_border = False
        h, w = mask_shape
        for y, x in sector_boundary_points:
            if y == 0 or y == h - 1 or x == 0 or x == w - 1:
                touches_border = True
                break

        return ExtentMetrics(
            midpoint=midpoint_dist,
            mean=mean_dist,
            max=max_dist,
            median=median_dist,
            std=std_dist,
            touches_border=touches_border,
        )



from __future__ import annotations

import functools
import logging
from pathlib import Path
import xml.etree.ElementTree as ElementTree

import imageio.v3 as imageio
import numpy as np
from skimage.util import img_as_ubyte

from eyepy.core.eyeenface import EyeEnface
from eyepy.core.eyemeta import EyeBscanMeta
from eyepy.core.eyemeta import EyeEnfaceMeta
from eyepy.core.eyemeta import EyeVolumeMeta
from eyepy.core.eyevolume import EyeVolume
from eyepy.io.utils import _compute_localizer_oct_transform
from eyepy.io.utils import _get_date_from_xml
from eyepy.io.utils import _get_first_as_float
from eyepy.io.utils import _get_first_as_int
from eyepy.io.utils import _get_first_as_str
from eyepy.io.utils import get_bscan_spacing

logger = logging.getLogger(__name__)

xml_format = {
    # Version
    'version': ('./SWVersion/Version', _get_first_as_str),
    # Number of A-Scans in each B-Scan, i.e. B-Scan width in pixel
    'size_x': (
        ".//ImageType[Type='OCT']../"
        'OphthalmicAcquisitionContext/Width',
        _get_first_as_int,
    ),
    # Number of B-Scans in OCT scan
    'n_bscans': (".//ImageType[Type='OCT']..", len),
    # Number of samples in an A-Scan, i.e. B-Scan height in pixel
    'size_y': (
        ".//ImageType[Type='OCT']../"
        'OphthalmicAcquisitionContext/Height',
        _get_first_as_int,
    ),
    # Width of a B-Scan pixel in mm
    'scale_x': (
        ".//ImageType[Type='OCT']../"
        'OphthalmicAcquisitionContext/ScaleX',
        _get_first_as_float,
    ),
    # Distance between two adjacent B-Scans in mm (Not given in the XMl we
    # could try to estimate it from the B-Scan positions)
    'distance': ('', lambda x: None),
    # Height of a B-Scan pixel in mm (In the XML the scale is given for
    # every B-Scan)
    'scale_y': (
        ".//ImageType[Type='OCT']../"
        'OphthalmicAcquisitionContext/ScaleY',
        _get_first_as_float,
    ),
    # Width of the SLO image in pixel
    'size_x_slo': (
        ".//ImageType[Type='LOCALIZER']../"
        'OphthalmicAcquisitionContext/Width',
        _get_first_as_int,
    ),
    # Height of the SLO image in pixel
    'size_y_slo': (
        ".//ImageType[Type='LOCALIZER']../"
        'OphthalmicAcquisitionContext/Height',
        _get_first_as_int,
    ),
    # Width of a pixel in the SLO image in mm
    'scale_x_slo': (
        ".//ImageType[Type='LOCALIZER']../"
        'OphthalmicAcquisitionContext/ScaleX',
        _get_first_as_float,
    ),
    # Height of a pixel in the SLO image in mm
    'scale_y_slo': (
        ".//ImageType[Type='LOCALIZER']../"
        'OphthalmicAcquisitionContext/ScaleY',
        _get_first_as_float,
    ),
    # Horizontal field size of the SLO image in dgr
    'field_size_slo': (
        ".//ImageType[Type='LOCALIZER']../"
        'OphthalmicAcquisitionContext/Angle',
        _get_first_as_int,
    ),
    # Scan focus in dpt
    'scan_focus': (
        ".//ImageType[Type='LOCALIZER']../"
        'OphthalmicAcquisitionContext/Focus',
        _get_first_as_float,
    ),
    # Examined eye (zero terminated string). "OS": Left eye; "OD": Right eye
    'scan_position': (
        './Patient/Study/Series/Laterality',
        lambda x: 'OD' if x[0].text == 'R' else 'OS',
    ),
    # Examination time
    'exam_time': ('./Patient/Study/Series/ReferenceSeries/ExamDate/Date',
                  _get_date_from_xml),
    # Scan pattern type: (Not given in XML)
    #   0 = Unknown pattern,
    #   1 = Single line scan (one B-Scan only),
    #   2 = Circular scan (one B-Scan only),
    #   3 = Volume scan in ART mode,
    #   4 = Fast volume scan,
    #   5 = Radial scan (aka. star pattern)
    'scan_pattern': ('', lambda x: 0),
    # Unique identifier of this OCT-scan (zero terminated string). This is
    # identical to the number <SerID> that is part of the file name.
    # Format: n[.m] n and m are numbers. The extension .m exists only for
    # ScanPattern 1 and 2. Examples: 2390, 3433.2
    'id': ('./Patient/Study/Series/ID', _get_first_as_int),
    # Unique identifier of the reference OCT-scan (zero terminated string).
    # Format: see ID, This ID is only present if the OCT-scan is part of a
    # progression otherwise this string is empty. For the reference scan of
    # a progression ID and ReferenceID are identical.
    'reference_id':
    ('./Patient/Study/Series/ReferenceSeries/ID', _get_first_as_int),
    # Internal patient ID used by HEYEX.
    'pid': ('./Patient/ID', _get_first_as_int),
    # User-defined patient ID (zero terminated string).
    'patient_id': ('./Patient/PatientID', _get_first_as_str),
    # First names of the patient
    'firstnames': ('./Patient/FirstNames', _get_first_as_str),
    # Last name of the patient
    'lastname': ('./Patient/LastName', _get_first_as_str),
    # Patient's date of birth
    'dob': ('./Patient/Birthdate/Date', _get_date_from_xml),
    # Internal visit ID used by HEYEX.
    'vid': ('', lambda x: None),
    # User-defined visit ID (zero terminated string). This ID can be defined
    # in the Comment-field of the Diagnosis-tab of the Examination Data
    # dialog box. The VisitID must be defined in the first row of the
    # comment field. It has to begin with an "#" and ends with any
    # white-space character. It can contain up to 23 alpha-numeric
    # characters (excluding the "#")
    'visit_id': ('', lambda x: None),
    # Date the visit took place. Identical to the date of an examination tab
    # in HEYEX.
    'visit_date': ('./Patient/Study/StudyDate/Date', _get_date_from_xml),
    # Type of grid used to derive thickness data. 0 No thickness data
    # available, >0 Type of grid used to derive thickness values. Thickness
    # data is only available for ScanPattern 3 and 4.
    'grid_type':
    ('./Patient/Study/Series/ThicknessGrid/Type', _get_first_as_int),
    # Internal progression ID (zero terminated string). All scans of the
    # same progression share this ID.
    'prog_id': ('./Patient/Study/Series/ProgID', _get_first_as_str),
}

xml_bscan_format = {
    # The same as the XML Version
    'version': ('./SWVersion/Version', _get_first_as_str),
    # X-Coordinate of the B-Scan's start point in mm.
    'start_x':
    ('./OphthalmicAcquisitionContext/Start/Coord/X', _get_first_as_float),
    # Y-Coordinate of the B-Scan's start point in mm.
    'start_y':
    ('./OphthalmicAcquisitionContext/Start/Coord/Y', _get_first_as_float),
    # X-Coordinate of the B-Scan's end point in mm. For circle scans, this
    # is the X-Coordinate of the circle's center point.
    'end_x':
    ('./OphthalmicAcquisitionContext/End/Coord/X', _get_first_as_float),
    # Y-Coordinate of the B-Scan's end point in mm. For circle scans, this
    # is the Y-Coordinate of the circle's center point.
    'end_y':
    ('./OphthalmicAcquisitionContext/End/Coord/Y', _get_first_as_float),
    # Number of segmentation vectors
    'num_seg': ('./Segmentation/NumSegmentations', _get_first_as_int),
    # Image quality measure. If this value does not exist, its value is set
    # to INVALID.
    'quality': ('./OphthalmicAcquisitionContext/ImageQuality',
                _get_first_as_float),
    # Horizontal shift (in # of A-Scans) of the classification band against
    # the segmentation lines (for circular scan only).
    'shift': ('', lambda x: None),
    # Intra volume transformation matrix. The values are only available for
    # volume and radial scans and if alignment is turned off, otherwise the
    # values are initialized to 0.
    'iv_trafo': ('', lambda x: None),
    # Bscan filename
    'scan_name': ('./ImageData/ExamURL', lambda x: x[0].text.split('\\')[-1]),
}


@functools.lru_cache(maxsize=4, typed=False)
def get_xml_root(filepath) -> ElementTree.Element:
    with open(filepath, encoding='utf-8', errors='replace') as mf:
        tree = ElementTree.parse(mf)
    return tree.getroot()


class HeXmlReader:

    def __init__(self, path):
        path = Path(path)
        if not path.suffix == '.xml':
            xmls = list(path.glob('*.xml'))
            if len(xmls) == 0:
                raise FileNotFoundError(
                    'There is no .xml file under the given filepath')
            elif len(xmls) > 1:
                raise ValueError(
                    'There is more than one .xml file in the given folder.')
            path = xmls[0]
        self.path = path
        self.xml_root: ElementTree.Element = get_xml_root(self.path)

        self.parsed_values = {
            name: func(self.xml_root[0].findall(xpath))
            for name, (xpath, func) in xml_format.items()
        }

    @property
    def bscan_meta(self) -> list[EyeBscanMeta]:
        bscan_meta = []
        for bscan in self.xml_root[0].findall(".//ImageType[Type='OCT'].."):
            parsed_values = {
                name: func(bscan.findall(xpath))
                for name, (xpath, func) in xml_bscan_format.items()
            }
            bscan_meta.append(
                EyeBscanMeta(
                    quality=parsed_values['quality'],
                    start_pos=(parsed_values['start_x'],
                               parsed_values['start_y']),
                    end_pos=(parsed_values['end_x'], parsed_values['end_y']),
                    pos_unit='mm',
                    scan_name=parsed_values['scan_name'],
                ))

        return bscan_meta

    @property
    def meta(self) -> EyeVolumeMeta:
        bscan_meta = self.bscan_meta
        return EyeVolumeMeta(
            scale_x=self.parsed_values['scale_x'],
            scale_y=self.parsed_values['scale_y'],
            scale_z=get_bscan_spacing(bscan_meta),
            scale_unit='mm',
            laterality=self.parsed_values['scan_position'],
            visit_date=self.parsed_values['visit_date'],
            exam_time=self.parsed_values['exam_time'],
            patient=self.patient,
            bscan_meta=bscan_meta,
            intensity_transform='default',
        )

    @property
    def patient(self):
        return dict(firstname=self.parsed_values['firstnames'],
                    lastname=self.parsed_values['lastname'],
                    dob=self.parsed_values['dob'],
                    pid=self.parsed_values['pid'],
                    patient_id=self.parsed_values['patient_id'])

    @property
    def localizer_meta(self) -> EyeEnfaceMeta:
        return EyeEnfaceMeta(
            scale_x=self.parsed_values['scale_x_slo'],
            scale_y=self.parsed_values['scale_y_slo'],
            scale_unit='mm',
            modality='NIR',
            laterality=self.parsed_values['scan_position'],
            field_size=self.parsed_values['field_size_slo'],
            scan_focus=self.parsed_values['scan_focus'],
            visit_date=self.parsed_values['visit_date'],
            exam_time=self.parsed_values['exam_time'],
        )

    @property
    def localizer(self) -> EyeEnface:
        localizer_pattern = ".//ImageType[Type='LOCALIZER']../ImageData/ExamURL"
        localizer_name = self.xml_root[0].find(localizer_pattern).text.split(
            '\\')[-1]
        localizer = imageio.imread(self.path.parent / localizer_name)
        if localizer.ndim == 3:
            localizer = img_as_ubyte(localizer[..., 0])
        else:
            localizer = img_as_ubyte(localizer)

        return EyeEnface(localizer, self.localizer_meta)

    @property
    def volume(self) -> EyeVolume:
        ## Check if scan pattern is supported by EyeVolume
        if self.parsed_values['scan_pattern'] == 2:
            msg = f'The EyeVolume object does not support scan pattern 2 (one Circular B-scan).'
            raise ValueError(msg)
        elif self.parsed_values['scan_pattern'] == 5:
            msg = f'The EyeVolume object does not support scan pattern 5 (Radial scan - star pattern).'
            raise ValueError(msg)

        bscans = []
        layer_heights = {}
        for index, bscan_root in enumerate(
                self.xml_root[0].findall(".//ImageType[Type='OCT']..")):
            scan_name = bscan_root.find('./ImageData/ExamURL').text.split(
                '\\')[-1]
            img = imageio.imread(self.path.parent / scan_name)
            if img.ndim == 3:
                img = img_as_ubyte(img[..., 0])
            else:
                img = img_as_ubyte(img)
            bscans.append(img)

            seglines = bscan_root.findall('.//SegLine')
            for segline in seglines:
                name = segline.find('./Name').text
                data = np.array(segline.find('./Array').text.split()).astype(
                    np.float32)
                data[data == 3.0e+38] = np.nan
                if name not in layer_heights:
                    layer_heights[name] = []
                layer_heights[name].append((index, data))

        data = np.stack(bscans, axis=0)

        layer_height_maps = {
            name: np.full((data.shape[0], data.shape[2]),
                          np.nan,
                          dtype=np.float32)
            for name in layer_heights
        }

        for name, heights in layer_heights.items():
            for index, layer_height in heights:
                layer_height_maps[name][index, :] = layer_height

        localizer = self.localizer
        volume_meta = self.meta
        transform = _compute_localizer_oct_transform(volume_meta,
                                                     localizer.meta,
                                                     data.shape)
        volume = EyeVolume(
            data=data,
            meta=volume_meta,
            localizer=localizer,
            transformation=transform,
        )

        for name, height_map in layer_height_maps.items():
            volume.add_layer_annotation(height_map, name=name)

        return volume
